package ri.analytics;

import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.UIManager;

import ri.util.AboutBox;
import ri.util.DialogSqliteConnect;
import ri.util.Emit;

/**Application to generate CSV files from database contents
 * Note: DO NOT index the 'name' column of mmw.sqlite narrow table. That slows it down by a factor of 15X
 *       Only needs indexes on entityId and value
 * */
public class StdfToCsv extends JFrame {

  private final static String _appVersionNo="1.0";
  private JTextField _tfldDbasename = new JTextField("mmw.sqlite", 30);
  private JTextField _tfldQuery = new JTextField("select distinct value from narrow where name='TEST_TEXT'", 40); //debug: sample queries
  private JTextField _tfldConstraints = new JTextField("", 30); //constraints on selected data
  private JTextField _tfldOutfile = new JTextField("out.csv", 30); //name of output CSV file
  private JTextField _tfldLimit = new JTextField("300", 5); //max number of records to emit to output csv file
  private JTextField _tfldProcessedCnt = new JTextField("0", 5);
  private JCheckBox _chkShow = new JCheckBox("Show", true); //shows intermediate results when checked
  
  private StdfToCsv _me=this; //convenience thing
  private volatile boolean _keepGoing=true; //used to stop long queries
  
  private Connection _inConnection = null; //connection to the input (customer data, narrow) database
  private Connection _outConnection = null; //connection to the output (temporary, wide) database
  private Statement _inStatement; //statement to (re)use for querying of the input database
  private Statement _outStatement; //statement to (re)use for querying of the output database
  private BufferedWriter _dout=null; //file handle for CSV output data file
  
  private ArrayList<String> _testnames = null; //lst of test names found in the input database
  private HashMap<String,HashMap<Integer,String>> _widemap = new HashMap<String,HashMap<Integer,String>>(); //<Testname, <DeviceNo, Value>>
  

  /**Constr*/
  public StdfToCsv() {
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    guiInit();

    //Set main window position and size:
    int xloc=50;
    int yloc=50;
    setLocation(xloc, yloc);
    Dimension scrnSz=Toolkit.getDefaultToolkit().getScreenSize();

    setSize(1000, ((scrnSz.height - yloc) / 9) * 8);
    validate();
    setVisible(true);
  }

  /**Initialize the gui components. Called by constr*/
  private void guiInit() {
    setTitle("STDF to CSV");

    Box pnlN=new Box(BoxLayout.Y_AXIS);
    
    JPanel pnlN1 = new JPanel();
    JButton butnDbaseSelect = new JButton("Select");
    JButton butnDbaseOpen = new JButton("Open");
    JButton butnDbaseClose = new JButton("Close");
    pnlN1.add(new JLabel("Database name:"));
    pnlN1.add(_tfldDbasename);
    pnlN1.add(butnDbaseSelect);
    pnlN1.add(butnDbaseOpen);
    pnlN1.add(butnDbaseClose);
    pnlN.add(pnlN1);

    JPanel pnl;
    pnl = new JPanel();
    pnl.add(new JLabel("Query:"));
    pnl.add(_tfldQuery);
    pnlN.add(pnl);

    pnl = new JPanel();
    pnl.add(new JLabel("Constraints:"));
    pnl.add(_tfldConstraints);
    pnlN.add(pnl);

    pnl = new JPanel();
    JButton butnOutfileSelect = new JButton("Select");
    pnl.add(new JLabel("Output (CSV) file:"));
    pnl.add(_tfldOutfile);
    pnl.add(butnOutfileSelect);
    pnlN.add(pnl);
    
    pnl = new JPanel();
    pnl.add(new JLabel("Records Limit:"));
    pnl.add(_tfldLimit);
    pnlN.add(pnl);

    pnl = new JPanel();
    JButton butnGo = new JButton("GO");
    JButton butnStop = new JButton("Stop");
    pnl.add(butnGo);
    pnl.add(butnStop);
    pnl.add(new JLabel("Count:"));
    pnl.add(_chkShow);
    pnl.add(_tfldProcessedCnt);
    pnlN.add(pnl);

    JPanel pnlS=new JPanel();
    
    butnDbaseSelect.setToolTipText("Select the source database");
    butnDbaseOpen.setToolTipText("Open the source database");
    butnDbaseClose.setToolTipText("Close the source database");
    butnOutfileSelect.setToolTipText("Select the output (CSV) file");

    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(pnlN, BorderLayout.NORTH);
    Component emit = Emit.checkoutTextArea(true);
    if(emit!=null) 
      getContentPane().add(emit, BorderLayout.CENTER);
    getContentPane().add(pnlS, BorderLayout.SOUTH);
    
    butnDbaseSelect.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doDbaseSelect(); }
    });
    butnDbaseOpen.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doOpen(); }
    });
    butnDbaseClose.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doClose(true); }
    });
    butnOutfileSelect.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doOutfileSelect(); }
    });
    butnGo.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doGo(); }
    });
    butnStop.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { _keepGoing=false; }
    });
  }
  
  /**Helps user select an input database*/
  private void doDbaseSelect() {
    //TODO: File chooser to help user choose a database file
  }

  /**Opens databases and files*/
  private void doOpen() {
    doClose(false); //just in case is currently open
    File dbFile = new File(_tfldDbasename.getText());
    try {
      String dbName = dbFile.getAbsolutePath();
      _inConnection = DialogSqliteConnect.doConnectTo(dbName, false);
      if(_inConnection==null)  
        Emit.out("No connection available to database '"+dbName+"'");
      else
        Emit.out("Input database '"+dbName+"' connected.");
      
      dbName = "tempwide.sqlite";
      _outConnection = DialogSqliteConnect.doConnectTo(dbName, false);
      if(_outConnection==null)  
        Emit.out("No connection available to temp database '"+dbName+"'");
      else
        Emit.out("Temp database '"+dbName+"' connected.");
    }
    catch(ClassNotFoundException  e) {
      Emit.out("Unable to connect to database because "+e);
    }
    catch( SQLException e) {
        Emit.out("Unable to connect to database because "+e);
      }
    
    try { _inStatement = _inConnection.createStatement(); } 
    catch(Exception e) { Emit.out("Unable to create generic 'statement' for the input database because: "+e); }
    try { _outStatement = _outConnection.createStatement(); } 
    catch(Exception e) { Emit.out("Unable to create generic 'statement' for the temp output database because: "+e); }
    
    //Open the disk output (CSV) file:
    try { if(_dout!=null) { _dout.close(); _dout=null; } } catch(Exception e) {} //close output file if is already open
    File f = new File(_tfldOutfile.getText());
    try { _dout = new BufferedWriter(new FileWriter(f)); }
    catch(Exception e) {
      Emit.out("Error opening the output (CSV) file: "+e);
    }
  }

  /**Closes databases and files*/
  private void doClose(boolean verbose) {
    try { _inStatement.close(); } catch(Exception e) {}
    try { _outStatement.close(); } catch(Exception e) {}

    if(_inConnection!=null) {
      try { 
        _inConnection.close(); 
        Emit.out("Input database closed.");
      }
      catch (SQLException e) {
        if(verbose)
          Emit.out("Error closing the input database connection: "+e);
      }
      _inConnection=null;
    }
    
    if(_outConnection!=null) {
      try { 
        _outConnection.close(); 
        Emit.out("Temp database closed.");
      }
      catch (SQLException e) {
        if(verbose)
          Emit.out("Error closing the temp database connection: "+e);
      }
      _outConnection=null;
    }
    
    //Close the output (CSV) file:
    if(_dout!=null) {
      try { _dout.close(); _dout=null; } catch(Exception e) {}
      Emit.out("Output (CSV) file closed.");
    }
  }
  
  /**Selects (CSV) output file*/
  private void doOutfileSelect() {
    //TODO: File chooser to help user choose an output filename
  }


  
  /**Issues the given SQL query. Assumes respCnt indicates the number of responses expected*/
  private ArrayList<String>[] getResultsFor(int respCnt, String sql) throws SQLException {
    ResultSet rs = _inStatement.executeQuery(sql);
    ArrayList<String>[] lists = new ArrayList[respCnt];
    for(int i=0; i<respCnt; i++)
      lists[i] = new ArrayList<String>();
    try {
      while(rs.next()) { //get contents of result set
        for(int i=0; i<respCnt; i++)
          lists[i].add(rs.getString(i+1)); //get each response
      }
    }
    finally { rs.close(); }
    return lists;
  }

  /**Utility to convert a list to a CSV string*/
  private String getCsvFor(ArrayList<String> list, boolean addQuotes) {
    StringBuffer sb = new StringBuffer();
    boolean first=true;
    for(String entry : list) {
      if(first) first=false; else sb.append(","); //comma sep on all but first
      if(addQuotes) sb.append("'").append(entry).append("'");
      else sb.append(entry);
    }
    return sb.toString();
  }

  
  /**Generates output (CSV) file from specified database and constraints*/
  private void doGo() {
    if(_inConnection==null)
      doOpen();
    if(_inConnection==null) {
      Emit.out("No database connection. Ignored");
      return;
    }
    
    try {
      String sql = "SELECT DISTINCT value FROM narrow WHERE name='TEST_TEXT'"; //get list of test names
      ArrayList<String>[] lists = getResultsFor(1, sql);
      _testnames = lists[0];
      Collections.sort(_testnames); //sort the testnames list
      String csvTnames = getCsvFor(_testnames,false);

      _dout.write("DEVNO,");
      _dout.write(csvTnames); //write test names line out to CSV file
      _dout.newLine();
      
      if(_chkShow.isSelected())
        Emit.out(csvTnames); //display the testnames found

      //Prep the temp output database:
      _outStatement.executeUpdate("DROP TABLE IF EXISTS wide"); //out with the old
      _outStatement.executeUpdate("CREATE TABLE IF NOT EXISTS wide ("+getCsvFor(_testnames, true)+");"); //create the new table
      
      //Do the time consuming queries outside of the Swing thread:
      new Thread() { 
        public void run() {
          try {
            HashSet<Integer> setDevnos = new HashSet<Integer>(); //Set to keep track of all device numbers found

            _keepGoing=true; //used to halt long term processes
            _tfldProcessedCnt.setText("0"); //display number of tests processed
            int cnt = _testnames.size();
            int i;
            for(i=0; i<cnt && _keepGoing; i++) { //for each testname found
              String testname = _testnames.get(i);
              
              HashMap<Integer,String> map = _widemap.get(testname); //temp map to hold the data so it can be emitted as wide
              if(map==null) {
                map = new HashMap<Integer,String>();
                _widemap.put(testname, map);
              }
              String sql = 
                  "SELECT n2.entityID, n2.value "+ //note: n1.value is test name
                  "FROM narrow n1, narrow n2 "+
                  "WHERE n1.entityID = n2.indexID AND n1.name='TEST_NAME' AND n1.value='"+testname+"' AND n2.name='R'"; 
              ArrayList<String>[] results = getResultsFor(2, sql);
              ArrayList<String> sdevnos = results[0];
              ArrayList<String> datavals = results[1];
              _tfldProcessedCnt.setText(Integer.toString(i+1));
              

              if(_chkShow.isSelected()) {
                String csvDevnos = getCsvFor(sdevnos, false);
                String csvValues = getCsvFor(datavals, false);
                Emit.out("DEVNUMS for "+testname+" = ["+csvDevnos+"]");
                Emit.out("VALS for "+testname+" = ["+csvValues+"]");
                Emit.out("");
              }
              for(String sdevno : sdevnos) //keep a list of all the device numbers
                setDevnos.add(Integer.parseInt(sdevno));
              
              //push this set of testname values into the wide table:
              //TODO: need to push this column in there vertically (at column 'testname')
//              sql = "INSERT INTO wide VALUES ("+csvValues+");"; <-Not this (horizontal)
//              _outStatement.executeUpdate(sql);
              
              int resultsLen = sdevnos.size();
              for(int j=0; j<resultsLen; j++) {
                String devno = sdevnos.get(j);
                String dataval = datavals.get(j);
                map.put(Integer.parseInt(devno), dataval);
              }
              //_tfldProcessedCnt.setText(Integer.toString(i+1)+".5"); //indicate output to 'temp' finished
            }
            if(i<cnt) Emit.out("Stopped");
            else Emit.out("Done"); //debug

            //Sort the device numbers:
            ArrayList<Integer> listDevnos = new ArrayList<Integer>(setDevnos);
            Collections.sort(listDevnos); //sort the device numbers
            Emit.out(""+listDevnos.size()+" distinct device numbers found"); //debug
            
            //Emit to the csv output file:
            int maxRecords = 300;
            try { maxRecords = Integer.parseInt(_tfldLimit.getText()); }
            catch(Exception e) {}
            _tfldLimit.setText(""+maxRecords); //update field with value actually being used (just in case)
            
            int cntr=0;
            try {
              for(Integer devno : listDevnos) {      //outer loop
                _dout.write(devno.toString());
                for(String testname : _testnames) { //inner loop
                  HashMap<Integer,String> mapVals = _widemap.get(testname); //all values for one test
                  if(mapVals!=null) {
                    String val = mapVals.get(devno);
                    if(val==null) val="";
                    _dout.write(","); //comma sep
                    _dout.write(val);
                  }
                }
                _dout.newLine();
                ++cntr;
                if(i % cntr == 100) Emit.outNoCr("."); //debug
                if(cntr >= maxRecords) break;
              }
              _dout.flush();
              Emit.out("\nCSV write done");
            }
            catch(Exception e) {
              Emit.out("Error writing to output CSV file: "+e);
            }
          }
          catch (SQLException e) {
            Emit.out("Error in SQL query: "+e);
          }
        }
      }.start(); 
    }
    catch(Exception e) { Emit.out("Error: "+e); }
  }
  
  /**Displays the Help | About Dialog*/
  private void doHelpAbout() {
    new AboutBox(this, new String[] { "STDF to CSV generator", "Version: " + _appVersionNo,
        "Copyright Roos Instruments (c) " + AboutBox.getYearStringFor(2013), "" }, null);
  }

  /**Overridden for exit on System Close*/
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if(e.getID()==WindowEvent.WINDOW_CLOSING) {
      doClose(false);
      System.exit(0);
    }
  }

  /**main method*/
  public static void main(String[] args) {
    try { UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); }
    catch (Exception e) {}
    new StdfToCsv();
  }

}
