package ri.stdf;

import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TimeZone;

import ri.util.ByteBuffer;
import ri.util.Emit;

/**
 * This class searches a supplied directory for files ending in .gat
 * with an expected format for guru attributes files.
 * An index database is also supplied with a schema of standard tuple
 * for each file not in the database (based on cid portion of file name)
 * the file is validated, parsed and the attributes entered into the
 * supplied database
 * some additional attributes are also added such as when imported
 * the file location fot the attributes and the body and whatever
 * @author markroos
 *
 */
public class MetadataIndexer {
  /**  Examines cid.gat files and imports into db if not present
  */
    Long sequence = 0L;
    Long entityId = 0L;
    public MetadataIndexer(){
      
    }
    public MetadataIndexer(File directory, Connection db){
      Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
      Calendar epoch = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
      epoch.set(Calendar.YEAR, 1990);
      epoch.set(Calendar.MONTH, 0);
      epoch.set(Calendar.DAY_OF_MONTH, 1);
      epoch.set(Calendar.HOUR_OF_DAY, 0);
      epoch.set(Calendar.MINUTE, 0);
      epoch.set(Calendar.SECOND, 0);
      Long delta = (Long) (cal.getTimeInMillis() - epoch.getTimeInMillis()) / 1000L;
      sequence = delta * 1000000L;
      ResultSet results = null;
      try {
        results=db.createStatement().executeQuery("select max(entityID) from narrow where name='ri.sys.Cid';");
        entityId =  results.getLong(1);
        System.out.println("last ID " + entityId);
        results.close();
      }
      catch (SQLException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
      indexMetadata(directory,db);
    }
  
    public void indexMetadata(File directory, Connection db){
      File[] list = directory.listFiles( new FilenameFilter(){
      public boolean accept( File dir, String name ) { 
        return name.toLowerCase().endsWith(".gat");
      }
      });
      int len = (list==null) ? 0 : list.length;
      if(len==0) return;
      for(int i=0; i<len; i++) {
        Emit.outNoCr(".");
        processFile(list[i], db);
      }
      Emit.out("");
    }

  private void processFile(File file, Connection db) {
    //Get files that are hash.gat

    //Update the db as required:
    try {
        updateAttributes(file, db); // insert into database
    }
    catch (SQLException e) {
      Emit.out("Error accessing dbase in "+getClass().getName()+".processFile(): "+e);
    }
  }

  /**Returns true if an entry for the specified cid is present in the database
   * @throws SQLException */
  private boolean isCidExists(String cid, Connection db) throws SQLException {
    ResultSet rs=null;
    try {
      rs=db.createStatement().executeQuery("SELECT entityID FROM narrow WHERE value='" + cid + "'");
      return rs.next(); //returns true if there is at least one matching entry, false otherwise 
    }
    finally {
      try { if (rs != null) rs.close(); } catch (Exception e) {}
    }
  }

  private void updateAttributes(File file, Connection db) throws SQLException {
    byte[] attribs=getFile(file);
    HashMap<String, String> guruKeys=null;
    guruKeys=convertGatCborToMap(attribs, -1); //convert the attributes array into a hashmap
    if(guruKeys.isEmpty()) return;
    if(isCidExists((String) guruKeys.get("ri.sys.Cid"),db)) return;
    // add extra keys here
    guruKeys.put("attrPath", file.getAbsolutePath());
    guruKeys.put("objPath", file.getParent() + "/" + guruKeys.get("ri.sys.ObjHash"));
    
    Iterator<String> iter=guruKeys.keySet().iterator();
    String key;
    String val;
    String objHash = "";
    // seq,eid,iid,name,value,value2
    PreparedStatement narrowPrepStatement=db.prepareStatement("insert into narrow values (?,?,?,?,?,?);");
    entityId++;
    try {
      while (iter.hasNext()) {
        key=(String)iter.next(); //guru key name
        val=(String)guruKeys.get(key); //value for key
        if(key.equalsIgnoreCase("ri.sys.ObjHash")) objHash = val;
        if (key != null && val != null) {
          narrowPrepStatement.setLong(1, ++sequence);
          narrowPrepStatement.setLong(2, entityId); //cid
          narrowPrepStatement.setString(4, key); //gkey
          narrowPrepStatement.setString(5, val); //gval
          narrowPrepStatement.addBatch();
          narrowPrepStatement.executeBatch();
        }
      }
    }
      catch (SQLException e) {
        Emit.out("Error in "+getClass().getName()+".updateAttributes() performing index (narrow) table insert: "+e);
    }
  }
  
  public void validateMetadataFiles(File directory){
    // search for all .gat files and check hash and signing
    File[] list = directory.listFiles( new FilenameFilter(){
    public boolean accept( File dir, String name ) { 
      return name.toLowerCase().endsWith(".gat");
    }
    });
    int len = (list==null) ? 0 : list.length;
    if(len==0) return;
    for(int i=0; i<len; i++) {
      byte[] attribs=getFile(list[i]);
      HashMap<String, String> guruKeys=null;
      guruKeys=convertGatCborToMap(attribs, -1); //convert the attributes array into a hashmap
      if(guruKeys.isEmpty()){
        System.out.println("verify failed "+list[i].getAbsolutePath());
      }
    }
  }
  
  /**
   * Converts the old stype gat file to a map of key values
   * First byte is not 0xD8
   * @param attribs
   * @param length
   * @return
   */
  static HashMap<String, String> convertGatBytesToMap(byte[] attribs, int length) {

    HashMap<String, String> guruKeys=new HashMap<String, String>(32, 0.9f); //create an empty map and fill it. note: average number of keys in an entry is 162)
    String key=null;
    StringBuffer sb=new StringBuffer();
    if (length < 0) length=attribs == null ? 0 : attribs.length;
    for (int i=0; i < length; i++) {
      byte b=attribs[i];

      if (b == '=' && key == null) { //do the 'key' portion (only take the first '=' any subsequent ones are part of the value)
        key=sb.toString();
        sb.setLength(0); //clear it out for next time
      }

      else if (b == 0x1c || b == 0x1d || (i == length - 1)) { //do the 'value' portion: on separators and at the end
        if (i == length - 1) sb.append((char)b); //grab that last char
        String val=sb.toString();
        if (key != null && key.length() > 0) { //if there is no key then we don't care
          if (val == null) val="";
          guruKeys.put(key, val);
        }
        sb.setLength(0); //clear it out for next time
        key=null;
      }

      else sb.append((char)b);
    }
    return guruKeys;
  }
  
  /**
   * Converts the new CBOR type gat file to a map of key values
   * First byte is 0xD8
   * Combine the various maps, Identity, attributes, sign
   * @param attribs
   * @param length
   * @return
   */
  static HashMap<String, String> convertGatCborToMap(byte[] attribs, int length) {
    int tmp = ((int)attribs[0]) & 0xff;
    if((((int)attribs[0]) & 0x80) != 0x80){ // check if old style
      return convertGatBytesToMap(attribs, length);
    }
    HashMap<String, String> guruKeys=new HashMap<String, String>(32, 0.9f); //create an empty map and fill it. note: average number of keys in an entry is 16
    ByteBuffer buffer = new ByteBuffer(attribs);
    while( buffer.atEnd() == false){
      int tag = buffer.nextTag();
      // check to do verify before security keys
      if(tag == 134){
        if(!RitdbIdentityGenerator.verifyGat(buffer, guruKeys)){
          System.out.println("verify failed ");
        }
      }else{
        buffer.fillMap(guruKeys);
      }
    }
    return guruKeys;
  }

  /** Helper get get bytes from file
   * 
   */
  static byte[] getFile(File file) {
    if (file == null) return null;
    if (!file.exists()) return null; //no such file found

    InputStream din=null;
    try {
      int flen=(int)file.length();
      din=new FileInputStream(file);
      byte[] bytes=new byte[flen];
      int cnt=din.read(bytes);
      if (cnt != flen) System.out.println("Warning: read " + cnt + " bytes but expected " + flen); //debug
      return bytes;
    }
    catch (Exception e) { return null; }
    finally {
      try { din.close(); } catch (Exception e) {}
    }
  }
  
  public static void main(String[] args) {
    System.out.println("Test of the indexer");
    HashMap<String, String> guruKeys1=new HashMap<String, String>(32, 0.9f);
    guruKeys1.put("Cid", "foobar");
    guruKeys1.put("Obsoleted", "false");   
    guruKeys1.put("ObjClass", "RitdbLog");
    ByteBuffer buffer = new ByteBuffer(10);
    buffer.writeTag(129);
    buffer.writeHashMap(guruKeys1);    
    HashMap<String, String> guruKeys=new HashMap<String, String>(32, 0.9f);
    guruKeys.put("ri.sys.Cid", "foobar");
    guruKeys.put("ri.sys.Obsoleted", "false");   
    guruKeys.put("ri.sys.ObjFormat", "ritdb");
    guruKeys.put("ri.sys.ObjClass", "RitdbLog");
    buffer.writeTag(130);
    buffer.writeHashMap(guruKeys);
    
    ByteBuffer buffer2 = new ByteBuffer(buffer.toBytes());
    HashMap<String, String> guruKeys2=new HashMap<String, String>(32, 0.9f);
    HashMap<String, String> identity=new HashMap<String, String>(32, 0.9f);
    int tag = buffer2.nextTag();
    buffer2.fillMap(identity);
    tag = buffer2.nextTag();
    buffer2.fillMap(guruKeys2);
    System.out.println(guruKeys2);
    System.out.println("Done.");
  }

}
