package ri.stdf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Scanner;
import java.util.TimeZone;

import ri.util.ByteBuffer;

/**
 * Generates an unique ID for this location and stores in in a 
 * disk file called gurukey.dat
 * @author markroos
 *
 */
public class RitdbIdentityGenerator {
  private static String guruId = null;
  private static BigInteger privateKeyInt = null;
  private static BigInteger publicKeyInt = null;
  private static BigInteger keyModulasInt = null;
  private static volatile Long lastTimeStamp = 0L;
  private static volatile int lastSequence = 0;
  private static final Object timeLock = new Object();
  
  public static String getGuruId(){
    if(!keyfileExists()){
      generateNewKeyFile();
    }else{
      readKeyFile();
    }
    return guruId;
  }
  /**given a byte buffer with the attributes and the object hash compute
   * and add the security section.
   * Uses CBOR encoding
   */
  public static void addSecurityAttributes(ByteBuffer buffer, byte[]objHash){
    //TODO
    buffer.writeTag(134); // add the tag
    buffer.writeArraySize(4);
    // compute the attrHash to this point
    byte[] attrHash = guruHash(buffer.toBytes());
    //System.out.println("orig hash  " + hexIt(attrHash));
    //System.out.println("attr size "+attribs.toBytes().length);
    // add the required security parameters of: attrHash,objHash,signer,sign
    //HashMap<String, String> signature=new HashMap<String, String>(32, 0.9f); // holds the signature attributes
    //AttrHash= hex SHA1 of the parameters up to (not including) the 16r1C separator of prior parameter
    //signature.put("AttrHash" , hexIt(attrHash));
    buffer.writeBytes(attrHash);
    //ObjHash=hex SHA1 of the object portion as transferred
    //signature.put("ObjHash" , hexIt(objHash));
    buffer.writeBytes(objHash);
    //SIGNER=guruID of the guru who signed the object,  local.ri.sys.Signer.
    //signature.put("SIGNER" , guruId);
    buffer.writeString(guruId);
    //SIGN=concatenation of the attrHash+ObjHash encrypted by the private key of the signer and converted to modPowerHex
    ByteBuffer tBuf = new ByteBuffer();
    tBuf.append(attrHash);
    tBuf.append(objHash);
    //signature.put("SIGN" ,encrypt(reverse(tBuf.toBytes())));
    buffer.writeBytes(encrypt(tBuf.toBytes()));
    System.out.println("orig sign  " + hexIt(tBuf.toBytes()));
    //buffer.writeHashMap(signature);  // append the map
  }
  
  /**
   * Checks the provided gat file for integrity
   * @param gat
   * @return
   */
  public static boolean verifyGat(ByteBuffer gat, HashMap guruKeys){
    //buffer is just past the tag for the signature section.
    int tmp = ((int)gat.get(0)) & 0x80;
    if(tmp != 0x80){ // check if old style
      return verifyGatLegacy(gat.toBytes());
    }
    // skip array size
    gat.nextCborSize();
    byte[] attrHash = guruHash(gat.toBytes());
    //System.out.println("read attrs  " + hexIt(attrHash));
    // extract the signer and sign information
    // skip the attrHash and the objHash
    gat.nextBytes();
    gat.nextBytes();
    String signer = gat.nextString();
    byte[] sign = gat.nextBytes();
    // decrypt the sign and compare with the attr hash
    byte[] dsign = decryptGat(sign);
    //System.out.println("decrypt  " + hexIt(dsign));
    // expecting 40 bytes here first is the attributes
    byte[] aHashBytes = Arrays.copyOfRange(dsign, 0, 20);
    if(Arrays.equals(aHashBytes, attrHash)) return true;  
    return false;
  }
  
  /**
   * Checks the provided gat file for integrity
   * @param gat
   * @return
   */
  public static boolean verifyGatLegacy(byte[] gat){
    //get the attributes bytes from the gat ( find last 1c )
    int pos = 0;
    for( int i = 0; i < gat.length; i++){
      if(gat[i] == 28) pos = i;
    }
    System.out.println(" end of attribs " + pos);
    byte[] attrHash = guruHash(Arrays.copyOfRange(gat, 0, pos));
    System.out.println("attr size "+ (Arrays.copyOfRange(gat, 0, pos).length));
    // extract the signer and sign information
    String signer = findValueForKey(gat, "SIGNER");
    String sign = findValueForKey(gat, "SIGN");
    // decrypt the sign and compare with the attr hash
    byte[] dsign = decryptGatLegacy(sign);
    // expecting 40 bytes here but could have extra leading 0 due to 
    // how big ints work
    byte[] aHashBytes = Arrays.copyOfRange(dsign, 0, 20);
    if(Arrays.equals(aHashBytes, attrHash)) return true;  
    return false;
  }
  
  /**
   * Takes a byte array containing the attribute hash and the object hash
   * and returns an hex string in mod power format
   * @param data
   */
  public static byte[] encrypt(byte[] data){
    // use sign magnitude conversion as must be positive int
    // removes extra leading zero if all zeros
    byte[] tmp = (new BigInteger(1, data).modPow(privateKeyInt, keyModulasInt)).toByteArray();
    byte[] ba = null;
    if(tmp[0] == 0){
      ba = Arrays.copyOfRange(tmp, 1, tmp.length);
    }else{
      ba = tmp;
    }
    return ba;
  }
  
  /**
   * Takes a byte array and returns the byte array for .gat files
   * @param data
   */
  public static byte[] decrypt(byte[] data){
    byte[] tmp = (new BigInteger(1, data)).modPow(publicKeyInt, keyModulasInt).toByteArray();
    // can have an extra leading 0 for sign bit which we do not want
    if(tmp[0] == 0){
      tmp =  Arrays.copyOfRange(tmp, 1, tmp.length);
    }
    return tmp;
  }
  
  /**
   * Takes a byte array and returns the byte[40] for .gat files
   * @param data
   */
  public static byte[] decryptGat(byte[] data){
    byte[] tmp = (new BigInteger(1, data)).modPow(publicKeyInt, keyModulasInt).toByteArray();
    // could be of a different size than expected due to extra or missing leading zeros
    if(tmp.length < 40){
      byte[] newTmp = new byte[40];
      System.arraycopy(tmp, 0, newTmp, 40 - tmp.length , tmp.length);
      return newTmp;
    }
    if(tmp.length > 40){
      byte[] newTmp =  Arrays.copyOfRange(tmp, tmp.length - 40, tmp.length);
      return newTmp;
    }
    return tmp; 
  }
  
  public static byte[] decryptGatLegacy(String data){
    // data is in hex mod power format with 4 leading size chars
    byte[] tmp = (fromHexString(data)).modPow(publicKeyInt, keyModulasInt).toByteArray();
    // could be of a different size than expected due to extra or missing leading zeros
    if(tmp.length < 40){
      byte[] newTmp = new byte[40];
      System.arraycopy(tmp, 0, newTmp, 40 - tmp.length , tmp.length);
      return reverse(newTmp);
    }
    if(tmp.length > 40){
      byte[] newTmp =  Arrays.copyOfRange(tmp, tmp.length - 40, tmp.length);
      return reverse(newTmp);
    }
    return reverse(tmp); 
  }
  
  private static boolean keyfileExists(){
   return new File("gurukey.dat").exists();
  }
  
  private static void generateNewKeyFile(){   
      // 8 chars X0000000 where the time is used to
      // generate a base 36 replacement for the zeros
    System.out.println("creating new identity");
      Long rawTime = (System.nanoTime() / 1000000L) & 0x7ffffffffL;
      guruId = "0000000";
      try {
        byte[] tmp = convertBase36(rawTime);
        tmp[0] = 88;
        guruId=new String(tmp, "US-ASCII"); 
      }
      catch (UnsupportedEncodingException e) {
      }
    getRSA();
    ByteBuffer bb=new ByteBuffer();
    // time stamp
    bb.append(Arrays.copyOfRange(convertBase36(guruTime()), 2, 8)); 
    bb.eol();
    bb.appendLine(":INFO"); 
    bb.appendLine("serialNumber=" + guruId); 
    if(privateKeyInt != null){
      bb.appendLine("CERT=" + toModPowHex(keyModulasInt) + toModPowHex(publicKeyInt) + guruId); // self signed so end in my guruId
      bb.appendLine("KEY=" + toModPowHex(privateKeyInt));
    }
    writeCff(bb,"gurukey.dat");
  }
  
  private static void readKeyFile(){
    File file = new File("gurukey.dat");
    Scanner din = null;
    try {
      din=new Scanner(file);
      while (din.hasNext()){
        String[] next = din.nextLine().split("=");
        if(next[0].equalsIgnoreCase("serialnumber")) guruId = next[1];
        if(next[0].equalsIgnoreCase("key")){
          int len = (Integer.parseInt(next[1].substring(0, 2), 16) * 4 + 4);
          String key = next[1].substring(0, len);
          privateKeyInt = fromHexString(key);
        }
        if(next[0].equalsIgnoreCase("cert")){ 
          int len1 = (Integer.parseInt(next[1].substring(0, 2), 16) * 4 + 4);
          int len2 = (Integer.parseInt(next[1].substring(len1, len1 + 2), 16) * 4 + 4);
          String riMod = next[1].substring(0, len1);
          String riPublicKey = next[1].substring(len1, len1 + len2);
          keyModulasInt = fromHexString(riMod);
          publicKeyInt = fromHexString(riPublicKey);
        }
      }
    }
    catch (FileNotFoundException e) {
    }
    finally{
      din.close();
    }
  }
  public static byte[] convertBase36(Long time) {
    //helper method to convert long to 8 byte array
    byte[] rtn = new byte[8];
    long result;
    long value = time;
    int pos = 7;
    for (int i = 0; i < 8; i++) {
      result = (value % 36) + 48;
      if (result >= 58) {
        result = result + 7;
      }
      rtn[pos] = (byte) result;
      value = value / 36;
      pos = pos - 1;
    }
    return rtn;
  }
  
  public static Long guruTime() {
    // helper method return guru time in seconds ( gmt offset from
    // 1/1/1990 )
    Calendar cal = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    Calendar epoch = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
    epoch.set(Calendar.YEAR, 1990);
    epoch.set(Calendar.MONTH, 0);
    epoch.set(Calendar.DAY_OF_MONTH, 1);
    epoch.set(Calendar.HOUR_OF_DAY, 0);
    epoch.set(Calendar.MINUTE, 0);
    epoch.set(Calendar.SECOND, 0);
    long delta = cal.getTimeInMillis() - epoch.getTimeInMillis();
    return (delta / 1000);
  }
  
  public static String nextCid(){
    // return an identifier for a log
    if(guruId == null) getGuruId();
    String time = "00000000";
    try {
      time=new String(nextGuruTimeStamp(), "US-ASCII");
    }
    catch (UnsupportedEncodingException e) {
    }
    return guruId + time;
  }
  
  private static void writeCff( ByteBuffer bytes, String file){
    int cks = 0; // checksum
    byte[] attrBytes=bytes.toBytes(); 
    for( int i = 0; i < attrBytes.length; i++) cks = cks + attrBytes[i];
    // now write out the file
    File f = new File(file);
    FileOutputStream dout=null;
    try {
      dout=new FileOutputStream(f);
      dout.write(attrBytes);
      dout.write(":CHECKSUM".getBytes("US-ASCII"));
      dout.write(0xd);
      dout.write(0xa);
      // write the check sum and the ending characters
      // update checksum for lf and :CHECKSUM 
      cks = cks + 653 + 23;
      // bytes for total checksum
      byte[] total = ("" + cks).getBytes("US-ASCII");
      for( int i = 0; i < total.length; i++) cks = cks + total[i];
      dout.write(total);
      dout.write(0xd);
      dout.write(0xa);
      cks = cks + 23;
      // now the extra ending characters to make total cks = 0
      int tmp = 256 - (cks % 256) + 16;
      for (int i = 0; i < 5; i++) {
        if (tmp > 74) {
          dout.write((char) 122);
          tmp -= 74;
        }
        else {
          if (tmp == 0) dout.write((char) 48);
          else {
            dout.write((char) (48 + tmp));
            tmp = 0;
          }
        }
      }
    }
    catch (IOException e) {
    }
    finally {
      try { dout.close(); } catch (Exception e) {}
    }
  }
  
  private static void getRSA(){
    // uses the java code to make a RSA key
    KeyPairGenerator kpg = null;
    try {
      kpg=KeyPairGenerator.getInstance("RSA");
    }
    catch (NoSuchAlgorithmException e) {
      System.out.println("Missing RSA provider");
    }
    kpg.initialize(512);
    KeyPair kp = kpg.genKeyPair();
    RSAPublicKey publicKey = (RSAPublicKey)kp.getPublic();
    RSAPrivateKey privateKey = (RSAPrivateKey)kp.getPrivate();
    privateKeyInt = privateKey.getPrivateExponent();
    publicKeyInt = publicKey.getPublicExponent();
    keyModulasInt = publicKey.getModulus();
  }
  /**Generates a hex string from a byte array where the
   * byte array is lsword ( high byte first) first and has the size in words
   * prepended also lsWord first.  The resulting byte array
   * is converted to a hex string
   * */
  private  static String toModPowHex(BigInteger key) {
    //TODO remove leading zero bytes from key
    byte[] tmp = key.toByteArray();
    byte[] ba = null;
    if(tmp[0] == 0){
      ba = Arrays.copyOfRange(tmp, 1, tmp.length);
    }else{
      ba = tmp;
    }
    int len=(ba == null) ? 0 : ba.length;  // length of the source
    int newLen=((len + 1) / 2) * 2;   // convert to an even number of words
    byte[] b=new byte[newLen];
    // lsb of the key goes to first location of the new array
    int j=len - 1;  // length of the source so point to the msbyte
    for (int i=0; i < len; i++, j--)
      b[i]=ba[j];
    StringBuffer sb=new StringBuffer();
    // first add two bytes of size, lsbyte first
    addaByteTo(sb, (byte)(newLen / 2 & 0xff)); //LSB
    addaByteTo(sb, (byte)((newLen / 2 >>> 8) & 0xff)); //MSB
    // now add the reversed key
    for (int i=0; i < newLen; i++)
      addaByteTo(sb, b[i]);
    return sb.toString();
  }
  
  private static void addaByteTo(StringBuffer sb, byte b) {
    // adds the high byte first
    int hi=(b & 0xf0) >>> 4;
    int lo=(b & 0xf);
    sb.append((char)((hi <= 9) ? '0' + hi : 'A' + hi - 10));
    sb.append((char)((lo <= 9) ? '0' + lo : 'A' + lo - 10));
  }
  
  /**Return the given byte array as a string of (upper case) hex bytes
   * If addSpaces is true then each byte is separated by a space*/
  public static String hexIt(byte[] ba) {
    int len = ba==null? 0 : ba.length;
    if(len<1) return "0";
    StringBuffer sb=new StringBuffer();
    try {
      for (int i=0; i < len; i++) {
        String s=Integer.toHexString(ba[i] & 0xff);
        if (s.length() <= 1) sb.append('0'); //for leading zero suppression
        sb.append(s.toUpperCase());
      }
    }
    catch (Exception e) {
      System.out.println("Error in RiHexString.hexIt(): " + e);
    }
    return sb.toString();
  }
  
  /**
   * the source string is a hex encoded byte array lsbyte first
   * with 4 leading size chars
   * @param s
   * @param extraBytes
   * @return
   */
  private static BigInteger fromHexString(String s) {
    int len=s.length();
    byte[] ba=new byte[(len - 4) / 2 ];
    int j=ba.length - 1;
    for(int i=4; i < len; i+=2)
      ba[j--]=(byte)Integer.valueOf(s.substring(i, i + 2), 16).intValue();
    return new BigInteger(1, ba);
  }
  
  /**Return the guru hash bytes for bytes*/
  public static byte[] guruHash(byte[] bytes) {
    try { 
      MessageDigest md=MessageDigest.getInstance("SHA"); 
      byte[] bytesHash=new byte[20];
      return md.digest(bytes);
    } 
    catch (Exception e) { return null; } //this would only happen if "SHA" was no longer a known thing
  }
  
  public static String guruHashString(byte[] bytes){
    return hexIt(guruHash(bytes));
  }
  
  /**
   * given a byte[] and a key to search for where the keyVals are
   * separated by 0x1d.  Return the value or null. Case sensitive
   * @param bytes
   * @param keyString
   * @return
   */
  public static String findValueForKey( byte[] bytes, String keyString){
    byte[] key = null;
    String tKey = "";
    int valStart = 0;
    try {key = keyString.getBytes("US-ASCII");}
    catch (UnsupportedEncodingException e) {};
    byte tChar = 0;
    byte cTest = 0;  // lowercase byte for first char in key
    if(key[0] > 96){
      cTest = (byte)( key[0] - 32);
    }else{
      cTest = key[0];
    }
    int i = 0;
    // move i so first byte is not 1c/1d
    while ((bytes[i] == 0x1c) || (bytes[i] == 0x1d) ) i++;
    boolean loop = true;
    while(i < bytes.length){
      // test the first byte for upper lower case compare
      if(bytes[i] > 96){ 
        tChar = (byte)(bytes[i] - 32);
      }else{
        tChar = bytes[i];
      }
      if((tChar == cTest) && bytes[i + key.length] == 61){  // look for equals sign
        // now covert to strings so we can do a case insensitive compare
        try {tKey = new String(Arrays.copyOfRange(bytes, i, i + key.length), "US-ASCII");}
        catch (UnsupportedEncodingException e) {};
        if( keyString.equalsIgnoreCase(tKey)){
          valStart = i + key.length + 1;
          i = valStart;
          // find the end of the value
          while (i < bytes.length){
            if((bytes[i] == 0x1d) || (bytes[i] == 0x1c) || (i == bytes.length - 1)){
              if(i == bytes.length - 1) i++;
              try {return new String(Arrays.copyOfRange(bytes, valStart, i), "US-ASCII");}
              catch (UnsupportedEncodingException e) {return null;}
            }
            i++;
          }
        }
    }
      // move past next 1d or 1c exit if too far
    i++;
    loop = true;
    while (loop){
      if(i > bytes.length - 1) return null;
      if( (bytes[i] == 0x1d) || (bytes[i] == 0x1c)){
        loop = false;
        i++;
      }else{
        i++;
      }
    }
    }
    return null;
   }
  
  private static byte[] reverse(byte[] array) {
    // in place reversal of bytes
    int i, j;
    byte temp;
    for (i = 0, j = array.length - 1; i < j; i++, j--) {
      temp = array[i];
      array[i] = array[j];
      array[j] = temp;
    }
    return array;
  }
  /**
   * generate a 8 byte ASCII encoded array with 6 chars of guru time in seconds
   * and 2 digits of sequence both encoded in base 36
   * @return
   */
  public static byte[] nextGuruTimeStamp() {
  // computes the next 8 byte timestamp,
  // 6 byte seconds since 1/1/90 base 36 + 2 base 36 digits for sequence
  // Guarantees monotonic stamps
  // gets current time and compares to lastTime, if > 1 sec reset
  // sequence and use that time, else bump sequence, if sequence
  // > 36 * 36 - 1 then bump time by 1 and reset sequence
  // convert result to base 36 bytes
    Long currentTime = guruTime();
    synchronized(timeLock){
      if (currentTime > lastTimeStamp) {
        lastTimeStamp = currentTime;
        lastSequence = 0;
      } else {
        lastSequence = lastSequence + 1;
        // check if overflow (2 digits of base36 => 36*36)
        if (lastSequence > 1295) {
          // if overflow, just shove over into the next time timeStamp
          lastTimeStamp = lastTimeStamp + 1;
          lastSequence = 0;
        }
      }
    }
    byte[] rtn = convertBase36((lastTimeStamp * 1296) + lastSequence);
    return rtn;
}
  
  private static void test(){
    byte[] hash = guruHash(new BigInteger("65537").toByteArray());
    System.out.println(new BigInteger(1, hash));
    System.out.println(encrypt(hash));
    System.out.println(new BigInteger(1,decrypt(encrypt(hash))));
  }
  
  public static void main(String[] args){
    // test a small conversion to from modPower format
    System.out.println(toModPowHex(new BigInteger("65537")));
    System.out.println(fromHexString((toModPowHex(new BigInteger("65537")))));
    // test encrypt / decrypt of RSA
    getRSA();
    test();
  }

}
