package ri.stdf;

import java.io.File;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

import ri.util.Emit;

public class RitdbIndexingThread extends Thread{ 
  private StdfUpload _parent; //parent window, for settings and display purposes
  
  /**Constr*/
  public RitdbIndexingThread(StdfUpload parent) {
    _parent = parent;
  }
  
  @Override
  public void run() {
    long t0 = System.currentTimeMillis();
    new RitdbMetadataCreator(new File(_parent.getOutDir())); // generate the files
    long t1 = System.currentTimeMillis();
    Emit.out("  Created Metadata ("+ ((t1-t0)/1000.0) +" sec)\n");
    Emit.out("Indexing metadata files");
    // need to open or create the indexDB
    Connection db =null;
    String dbName = "guruKeys.sqlite";
    try { db = DriverManager.getConnection("jdbc:sqlite:"+dbName); } //open connection to 'guruKeys' database
    catch(Exception e) { 
      Emit.out("Error. Unable to connect to database "+dbName+" because: "+e);
      return;
    } 
    try {
      db.createStatement().executeUpdate(
          "create table if not exists " + "narrow" + " ("+
          "sequence INTEGER PRIMARY KEY,"+
          "entityID INTEGER,"+
          "indexID INTEGER,"+
          "name TEXT,"+
          //"value TEXT,"+ //for postgres
          "value NONE,"+
          "value2 NONE);"
          );
    }
    catch (SQLException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    }
    try { 
      db.setAutoCommit(false);
      new MetadataIndexer(new File(_parent.getOutDir()), db); 
      db.commit(); //commit all those changes
      db.setAutoCommit(true);
    }
    catch (SQLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    finally { //ensures the database gets closed regardless of exit path
      try { db.close(); } catch(Exception e) {}
    }
    t0 = System.currentTimeMillis();
    Emit.out("  Finished Indexing Metadata ("+ ((t0-t1)/1000.0) +" sec)\n");

  }



}
