package ri.stdf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.UUID;

import ri.util.Emit;

import com.tragicphantom.stdf.Header;
import com.tragicphantom.stdf.Record;
import com.tragicphantom.stdf.RecordData;
import com.tragicphantom.stdf.RecordVisitor;
import com.tragicphantom.stdf.STDFReader;

public class Stdf4ToStdf5Translator implements RecordVisitor {
  public static int _gmtOffsetMin = -480; // offset in minutes to use for gmt time
  private String fileName = null;
  private STDFReader reader = null;
  public  int recordCnt = 0;
  private long  sequence = 0;  // holder for sequence override in secs
  private long  realSequence = 0;  // actual value for seq column gmt seconds * 100000 + fraction
  private int entityID = 1;  // runID is 1 first non run is 2
  private int eventGroup = 0; // set of events from a single touchdown
  private boolean inGroup = false; // true if between pirs and prrs, used to bump eventGrpoup
  private Connection _sqlConnection=null;     //database connection
  private Statement _sqlStatement=null;       //reusable statement for communicating with the database
  private PreparedStatement _stringInsert = null;
  private PreparedStatement _intInsert = null;
  private PreparedStatement _doubleInsert = null;
  private int runID = 0;      // entity ID for the run
  private int waferID = 0;      // entity ID for the current wafer
  private String waferIDvalue = null;      // value of ID for the current wafer
  private int[] currentTE = new int[256];  // entity ID per site for current test event, between PIR/PRR
  // now a map from the test number in stdf to the entityID of the test
  private Hashtable<Integer,Integer> testMap = new Hashtable<Integer,Integer>();
  // if test indexing is via the name and not the number then we need a name to entity map
  private Hashtable<String,Integer> testNameMap = new Hashtable<String,Integer>();
  // and to support the textTxt field changing from ptr to ptr we need to keep the last value per entityID
  private Hashtable<Integer,String[]> testTextMap = new Hashtable<Integer,String[]>();
  // and a map from test entityID to a some siteFlags if a PTR has been processed
  // site flags keep info on which sites have been processed
  private Hashtable<Integer,Long> testUpdate = new Hashtable<Integer,Long>();
  // and a map from test entityID to a true if a TSR has been processed
  private Hashtable<Integer,Boolean> testName = new Hashtable<Integer,Boolean>();
  // current limit file entityID.
  private int currentLimits = 0;
  private String _translationType=null; //a string that helps guide the handling of non-standard entities, e.g., "pure", "normal" 
  private boolean _testByNumber = true; // if false uses the testTxt field as the unique test ID
  private int dtrID = 0;
  private int bpsID = 0;
  private int epsID = 0;
  
  //@Override
  public void afterFile() {
    
  }

  //@Override
  public void beforeFile() {
  }
  
  /** constructor
   * 
   */
  public Stdf4ToStdf5Translator(File file, Connection sqlConnection, String translationType) throws SQLException, FileNotFoundException, IOException, ParseException {
    fileName = file.getName();
    _sqlConnection = sqlConnection;
    _translationType = translationType;
    if(_translationType.equalsIgnoreCase("2: testByName")) _testByNumber = false;
    _sqlStatement=_sqlConnection.createStatement(); //reusable statement for communicating with the database
    try { _sqlStatement.executeUpdate("PRAGMA synchronous = OFF;");
          _sqlStatement.executeUpdate("PRAGMA journal_mode = OFF;");} //generates an error in postgres but speeds things up in sqlite by allowing the database to perform ops without verifyng the disk writes went through
    catch(Exception e) { 
      //Emit.out("Warning: 'PRAGMA synchronous' not supported. The error was: [" + e + "]"); 
    }
    createTable("narrow", "narrow");
    _sqlConnection.setAutoCommit(false);
    runID = entityID;
    insertObject(sequence, entityID, 0, "ENTITY_TYPE", "RUN_INFO");
    insertObject(sequence, entityID, 0, "STDF_FILE", fileName);
    //insertObject(sequence, entityID, 0, "STDF_VER", "v5");
    insertObject(sequence, entityID, 0, "RUN_UUID", RitdbIdentityGenerator.nextCid());
    insertObject(sequence, entityID, 0, "RUN_GROUP_UUID", UUID.randomUUID().toString());
    reader=new STDFReader(file); //note: accepts File, InputStream or filename
    reader.parse(this); //as it parses, the reader calls handleRecord (below) for each entry
    _sqlConnection.commit(); //commit all those changes
    _sqlConnection.setAutoCommit(true);

    try { _sqlStatement.executeUpdate("PRAGMA synchronous = ON;"); } //restore setting allowing the database to verify writes to the file system
    catch(Exception e) {}
    try { _sqlStatement.executeUpdate("VACUUM"); } //restore setting allowing the database to verify writes to the file system
    catch(Exception e) {}

  }
  
  /**
   * Create the database and the single table
   * table is sequence int,entity int,index int, name string, value string | int | float
   * index is nullable, sequence is unique
   */
  private void createTable(String dName, String tName) {
    //Create database table if needed:
    String tableName = tName;
    //create sql table
    tableName=tName;
    try {
      _sqlStatement.executeUpdate(
          "create table if not exists " + tableName + " ("+
          "sequence INTEGER PRIMARY KEY,"+
          "entityID INTEGER,"+
          "indexID INTEGER,"+
          "name TEXT,"+
          //"value TEXT,"+ //for postgres
          "value NONE,"+
          "value2 TEXT);"
          );
    }
    catch (SQLException e) {
      e.printStackTrace();
    }   
    // now create the necessary prepared statements for each type
    //create a prepared sql statement that has the necessary fields for this stdf record type
    try {
      _stringInsert=_sqlConnection.prepareStatement("insert into " + tableName + " values (?,?,?,?,?,?);");
    }
    catch (SQLException e) {
      e.printStackTrace();
    } //prepared statement: "insert into table..."  
  }
  
  private void dropDataTable() {
    // drops the data
    try{
    Statement sqlStatement = _sqlConnection.createStatement(); 
    DatabaseMetaData dbmd = _sqlConnection.getMetaData(); 
    ResultSet rs = dbmd.getTables(_sqlConnection.getCatalog(), null, null, new String[]{"TABLE"});
    ArrayList list = new ArrayList();
    while(rs.next())
      list.add(rs.getString(3)); //table name is index #3
    rs.close();
    Iterator iter = list.iterator();
    while(iter.hasNext()) {
      String tablename = iter.next().toString();
      sqlStatement.executeUpdate("drop table "+tablename);
      Emit.out("Dropped table "+tablename);
    }
    sqlStatement.executeUpdate("VACUUM");
  }
  catch(Exception e) {
    System.out.println("Error dropping tables: "+e);
  }
  }
 
  /** two possible insert types,  no updates only inserts
   * 
   */
  private boolean insertObject(long seq, int eid, int index, String name, Object val) {
    //Get or create the prepared SQL statement:
    //Populate the data and send it to the database:
    // object can be string, long, double
    if(val == null) {
      return false;
    }
    // check if empty string and don't insert it
    if( val.getClass() == String.class) {
      String tString = (String) val;
      if(tString.trim().isEmpty()) {
        return false;
      }
    }
    ++realSequence;   
    PreparedStatement prep = _stringInsert;
    try{
      prep.setLong(1, realSequence);
      prep.setInt(2, eid);
      prep.setInt(3, index);
      prep.setString(4, name);

/*For sqlite:*/
      Class clazz = val.getClass();
      if(clazz == String.class)       prep.setString(5, (String)val);   
      else if(clazz == Long.class)    prep.setLong(5, (Long)val);   
      else if(clazz == Integer.class) prep.setInt(5, (Integer)val);   
      else if(clazz == Double.class)  prep.setDouble(5, (Double)val);   
      else if(clazz == Float.class)   prep.setFloat(5, (Float)val);   
      else if(val instanceof byte[])  prep.setBytes(5, (byte[])val);

/*For postgres:* /
      prep.setString(5, val==null ? null : val.toString());
      prep.setString(6, null);
/* */

      prep.addBatch();
      prep.executeBatch();
      }
    catch(Exception e) {
      String emsg = "Error in "+getClass().getName()+".insertObject(): "+e; 
      System.out.println(emsg);
      e.printStackTrace();
    }
    return true;
  }

  private boolean insertObject(long seq, int eid, int index, String name, Object val, String val2) {
    //Get or create the prepared SQL statement:
    //Populate the data and send it to the database:
    // object can be string, long, double
    PreparedStatement prep = _stringInsert;
    // sequence in seconds
    ++realSequence;      
    try{
      prep.setLong(1, realSequence);
      prep.setInt(2, eid);
      prep.setInt(3, index);
      prep.setString(4, name);
      prep.setString(6, val2);

/*For sqlite: */            
      if(val != null) {
        Class clazz = val.getClass();
        if(clazz == String.class)       prep.setString(5, (String)val);   
        else if(clazz == Long.class)    prep.setLong(5, (Long)val);   
        else if(clazz == Integer.class) prep.setInt(5, (Integer)val);   
        else if(clazz == Double.class)  prep.setDouble(5, (Double)val);   
        else if(clazz == Float.class)   prep.setFloat(5, (Float)val);   
        //if(clazz == byte[])           prep.setBytes(5, (byte[])val);}
      }
/* For postgres * /      
      prep.setString(5, val==null ? null : val.toString());
/* */

      prep.addBatch();
      prep.executeBatch(); 
      }
    catch(Exception e) {
      String emsg = "Error in "+getClass().getName()+".insertObject(): "+e; 
      System.out.println(emsg);
      e.printStackTrace();
    }
    return true;
  }
  
  /*private int getTestEntityFor(int testNumber) {
    if(testMap.containsKey(testNumber)) {
      return testMap.get(testNumber);
    }else{
      entityID++;
      testMap.put(testNumber, entityID);
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_INFO");
      insertObject(sequence, entityID, 0, "TEST_NUMBER", testNumber);  // put in the test number
      return entityID;
    }
  }
  */
  
  private int getTestEntityFor(int testNumber, String testText, Long site) {
    // depending on mode this handles the various uses of testTest/testName
    // TODO if _testByNumber is true then we use this to see if the testText has changed
    // since the last PTR,  if it has then we save it again as part of the result
    // if _testByNumber is false then we use testText as the unique test identifier
    // and use it to look up the entity ID
    int testEntityId ;
    if(_testByNumber == true){
      if(testMap.containsKey(testNumber)) {
        testEntityId =  testMap.get(testNumber);
      }else{
        entityID++;
        testMap.put(testNumber, entityID);
        insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_INFO");
        insertObject(sequence, entityID, 0, "TEST_NUMBER", testNumber);  // put in the test number
        insertObject(sequence, entityID, 0, "TEST_UNIQUE_ID", testNumber);  // put in the test number
        testEntityId = entityID;
      }
      return testEntityId;
    }else{
      if(testNameMap.containsKey(testText)) {
        return testNameMap.get(testText);
      }else{
        entityID++;
        testNameMap.put(testText, entityID);
        insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_INFO");
        insertObject(sequence, entityID, 0, "TEST_NUMBER", testNumber);  // put in the test number
        insertObject(sequence, entityID, 0, "TEST_NAME", testText);  // put in the test number
        insertObject(sequence, entityID, 0, "TEST_UNIQUE_ID", testText);  // put in the test number
        return entityID;
      }
    }
  }
  
  private void updateTextText(int testEntity, int site, String testText){
    // now update the testText
    if(_testByNumber == false) return;  // no test text if used for unique ID
    if(testText == null) return;
    if( site < 0 ) return;  // could be TSR so skip only for PTR,MPR and FTR
    //first update the test info section if this is the first time through
    if(testUpdate.containsKey(testEntity)) {
      Long test = testUpdate.get(testEntity);
      if(test == 0L) testTextMap.put(testEntity, new String[33]);  //TODO only 32 sites for now
      if((((int) (test >> site)) & 0x1) == 0){  // check to see if the flag is set "
        insertObject(sequence, testEntity, site, "TEST_TEXT", testText);
        test = testUpdate.get(testEntity);
        testUpdate.put(testEntity, (1L << site) + test);
        if(site < 32) (testTextMap.get(testEntity))[site] = testText;  // per site test text
      }
    }

    //TODO add support for where test_text is part of result per site
    if( site > 32){
      System.out.println("too may sites, max is 32"); 
      return;
    }
    if(!((testTextMap.get(testEntity))[site]).equalsIgnoreCase(testText)){
      int te = currentTE[site]; // per site test event
      insertObject(sequence, te, testEntity, "TEST_TEXT", testText);
    }
  }
  
  private String convertUnixTimeStamp(long timeSecs) {
    if(timeSecs == 0) {
      return null;
    }
    // create a calendar object and set the time
    Calendar cal = Calendar.getInstance();
    cal.setTimeInMillis(timeSecs * 1000);
    SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    return df.format(cal.getTime());
  }
  
  private String convertUnixTimeStampGMT(long timeSecs) {
    if(timeSecs == 0) {
      return null;
    }
    if( timeSecs * 1000000L > realSequence) {
      realSequence = timeSecs * 1000000L;
    }
    // create a calendar object and set the time
    Calendar cal = Calendar.getInstance();
    cal.setTimeInMillis(((_gmtOffsetMin * 60) + timeSecs) * 1000);
    SimpleDateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
    return df.format(cal.getTime());
  }
  
  public void handleUnknownRecord(Header header, byte[] bytes) {
    Emit.out("WARNING: Skipping unknown record type: " + header.getType() + ", " + header.getSubType() + " size " + header.getLength());
    try {
      _sqlConnection.commit();
    }
    catch (SQLException e) {
      e.printStackTrace();
    } //commit all those changes up  to here
  }
  
  //@Override
  public void handleRecord(Record record) {
    try { 
      ++recordCnt; //keep track of how many STDF records are processed
      if((recordCnt % 10000) == 0) {
        Emit.outNoCr(".");
      }
      RecordData recordData=record.getData(); //throws ParseException
      String recordType=recordData.getType(); //type of record (Mir, Ptr etc.)
      HashMap<String, Object> fields = recordData.getFields();
      //System.out.println("parsing "+ recordType);
      if(recordType.equals("Mir")) { this.insertMir(fields); }
      if(recordType.equals("Far")) { this.insertFar(fields); }
      if(recordType.equals("Mrr")) { this.insertMrr(fields); }
      if(recordType.equals("Tsr")) { this.insertTsr(fields); }
      if(recordType.equals("Pir")) { this.insertPir(fields); }
      if(recordType.equals("Prr")) { this.insertPrr(fields); }
      if(recordType.equals("Ptr")) { this.insertPtr(fields); }
      if(recordType.equals("Pcr")) { this.insertPcr(fields); }
      if(recordType.equals("Hbr")) { this.insertHbr(fields); }
      if(recordType.equals("Sbr")) { this.insertSbr(fields); }
      if(recordType.equals("Sdr")) { this.insertSdr(fields); }
      if(recordType.equals("Dtr")) { this.insertDtr(fields); }
      if(recordType.equals("Rdr")) { this.insertRdr(fields); }
      if(recordType.equals("Wir")) { this.insertWir(fields); }
      if(recordType.equals("Wcr")) { this.insertWcr(fields); }
      if(recordType.equals("Wrr")) { this.insertWrr(fields); }
      if(recordType.equals("Pgr")) { this.insertPgr(fields); }
      if(recordType.equals("Pmr")) { this.insertPmr(fields); }
      if(recordType.equals("Ftr")) { this.insertFtr(fields); }
      if(recordType.equals("Plr")) { this.insertPlr(fields); }
      if(recordType.equals("Mpr")) { this.insertMpr(fields); }
      if(recordType.equals("Gdr")) { this.insertGdr(fields); }
      if(recordType.equals("Bps")) { this.insertBps(fields); }
      if(recordType.equals("Eps")) { this.insertEps(fields); }
      if(recordType.equals("Atr")) { this.insertAtr(fields); }
    }
    catch(ParseException e) {
      String emsg = "Error in "+getClass().getName()+".handleRecord(): "+e; 
      Emit.out(emsg);
      System.out.print(emsg);
      //System.exit(1); //debug
      
    }
  }
  
  private void insertFar(HashMap  fields) {
    //insertObject(sequence, runID, 0, "CPU_TYPE", fields.get("CPU_TYPE"));
    //insertObject(sequence, runID, 0, "FILE_STDF_VER", fields.get("STDF_VER"));
      
  }
  
  /**  Handles MIR Record mapping
   * 
   */
  private void insertMir(HashMap  fields) {
    long time = (Long) fields.get("SETUP_T");
    if(time != 0) {
      insertObject(time, runID, 0, "JOB_SETUP_DTTM_UTC", convertUnixTimeStampGMT(time));
      insertObject(time, runID, 0, "JOB_SETUP_DTTM", convertUnixTimeStamp(time));
    }
    time = (Long) fields.get("START_T");
    if(time != 0) {
      insertObject(time, runID, 0, "RUN_START_DTTM_UTC", convertUnixTimeStampGMT(time));
      insertObject(time, runID, 0, "RUN_START_DTTM", convertUnixTimeStamp(time));
    }
    long tmp = (Long) fields.get("BURN_TIM");
    if(tmp != 65535L) {
        insertObject(sequence, runID, 0, "BURNIN_TIME", fields.get("BURN_TIM"));
    }
    insertObject(sequence, runID, 0, "TRANSLATOR_VERSION", "51"); 
    insertObject(sequence, runID, 0, "RITDB_VERSION", "0.2");
    insertObject(sequence, runID, 0, "NAMESPACE", "datalog");
    insertObject(sequence, runID, 0, "TESTER_ID", fields.get("NODE_NAM"));   
    insertObject(sequence, runID, 0, "LOT_ID", fields.get("LOT_ID"));
    insertObject(sequence, runID, 0, "STATION_NUMBER", fields.get("STAT_NUM"));
    insertObject(sequence, runID, 0, "TEST_MODE_CODE", fields.get("MODE_COD"));
    insertObject(sequence, runID, 0, "RETEST_CODE", fields.get("RTST_COD"));
    insertObject(sequence, runID, 0, "DATA_PROTECTION_CODE", fields.get("PROT_COD"));
    insertObject(sequence, runID, 0, "COMMAND_MODE_CODE", fields.get("CMOD_COD"));   
    insertObject(sequence, runID, 0, "TESTER_TYPE", fields.get("TSTR_TYP"));
    insertObject(sequence, runID, 0, "PRODUCT_ID", fields.get("PART_TYP"));
    insertObject(sequence, runID, 0, "JOB_NAME", fields.get("JOB_NAM"));    
    insertObject(sequence, runID, 0, "JOB_REVISION", fields.get("JOB_REV"));
    insertObject(sequence, runID, 0, "STEP_NAME", fields.get("SBLOT_ID"));
    insertObject(sequence, runID, 0, "OPERATOR_ID", fields.get("OPER_NAM"));    
    insertObject(sequence, runID, 0, "TESTER_EXEC_SW_TYPE", fields.get("EXEC_TYP"));
    insertObject(sequence, runID, 0, "TESTER_EXEC_SW_VER", fields.get("EXEC_VER"));
    insertObject(sequence, runID, 0, "TEST_PASS_NAME", fields.get("TEST_COD"));
    insertObject(sequence, runID, 0, "TEST_TEMPERATURE", fields.get("TST_TEMP"));
    insertObject(sequence, runID, 0, "USER_JOB_DESCRIPTION", fields.get("USER_TXT"));
    insertObject(sequence, runID, 0, "AUXILIARY_FILE", fields.get("AUX_FILE"));    
    insertObject(sequence, runID, 0, "PACKAGE_TYPE", fields.get("PKG_TYP"));
    insertObject(sequence, runID, 0, "PRODUCT_FAMLY_ID", fields.get("FAMLY_ID"));
    insertObject(sequence, runID, 0, "DATE_CODE", fields.get("DATE_COD"));
    insertObject(sequence, runID, 0, "TEST_FACILITY_ID", fields.get("FACIL_ID"));
    insertObject(sequence, runID, 0, "TEST_FLOOR_ID", fields.get("FLOOR_ID"));  
    insertObject(sequence, runID, 0, "FAB_PROCESS_ID", fields.get("PROC_ID"));
    //insertObject(sequence, runID, 0, "OPER_FRQ", fields.get("OPER_FRQ"));
    insertObject(sequence, runID, 0, "TEST_SPECIFICATION_NAME", fields.get("SPEC_NAM"));
    insertObject(sequence, runID, 0, "TEST_SPECIFICATION_VERSION", fields.get("SPEC_VER"));
    insertObject(sequence, runID, 0, "FLOW_ID", fields.get("FLOW_ID"));
    insertObject(sequence, runID, 0, "SETUP_ID", fields.get("SETUP_ID"));
    insertObject(sequence, runID, 0, "DESIGN_REV", fields.get("DSGN_REV"));
    insertObject(sequence, runID, 0, "ENGINEERING_LOT_ID", fields.get("ENG_ID"));
    insertObject(sequence, runID, 0, "ROM_CODE", fields.get("ROM_COD"));
    insertObject(sequence, runID, 0, "TESTER_SERIAL_NUMBER", fields.get("SERL_NUM"));
    insertObject(sequence, runID, 0, "SUPERVISOR_ID", fields.get("SUPR_NAM"));
    insertObject(sequence, runID, 0, "USER_LOT_DESCRIPTION", fields.get("USR_DESC"));
    insertObject(sequence, runID, 0, "LOT_DISPOSITION_CODE", fields.get("DISP_COD"));
    insertObject(sequence, runID, 0, "EXEC_LOT_DESCRIPTION", fields.get("EXC_DESC")); 

    }
  
    private void insertWir(HashMap  fields) {

      if(waferID == 0) {       
        entityID++;  // get the next entity ID
        waferID = entityID; 
        insertObject(sequence, waferID, 0, "ENTITY_TYPE", "WAFER_EVENT");
      }
      eventGroup = 0;
      waferIDvalue = (String) fields.get("WAFER_ID"); // insert at WRR

      Long time  = (Long) fields.get("START_T");
      insertObject(time, waferID, 0, "WAFER_START_DTTM", convertUnixTimeStamp(time));
      insertObject(time, waferID, 0, "WAFER_START_DTTM_UTC", convertUnixTimeStampGMT(time));
    }
  
    /**  Handles MRR Record mapping
     * 
     */
    private void insertMrr(HashMap  fields) {
      long time = (Long) fields.get("FINISH_T");
      insertObject(time, runID, 0, "RUN_END_DTTM", convertUnixTimeStamp(time));
      insertObject(time, runID, 0, "RUN_END_DTTM_UTC", convertUnixTimeStampGMT(time));
      insertObject(sequence, runID, 0, "LOT_DISPOSITION_CODE", fields.get("DISP_COD"));
      insertObject(sequence, runID, 0, "USER_LOT_DESCRIPTION", fields.get("USR_DESC"));
      insertObject(sequence, runID, 0, "EXEC_LOT_DESCRIPTION", fields.get("EXC_DESC"));
    }
    
    /**  Handles WRR Record mapping
     * 
     */
    private void insertWrr(HashMap  fields) {
      long time = (Long) fields.get("FINISH_T");
      insertObject(time, waferID, 0, "WAFER_STOP_DTTM", convertUnixTimeStamp(time));
      insertObject(time, waferID, 0, "WAFER_STOP_DTTM_UTC", convertUnixTimeStampGMT(time));
      insertObject(sequence, waferID, 0, "TOUCHDOWN_COUNT", eventGroup);
      insertObject(sequence, waferID, 0, "WAFER_FRAME_ID", fields.get("FRAME_ID"));
      if(fields.get("WAFER_ID") != null) waferIDvalue = (String) fields.get("WAFER_ID");
      insertObject(sequence, waferID, 0, "WAFER_ID", waferIDvalue);
      insertObject(sequence, waferID, 0, "WAFER_MASK_ID", fields.get("MASK_ID"));
      insertObject(sequence, waferID, 0, "WAFER_USER_DESC", fields.get("USR_DESC"));
      insertObject(sequence, waferID, 0, "WAFER_EXEC_DESC", fields.get("EXC_DESC"));
      // now add the WAFER_SUMMARY section
      entityID++;
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "WAFER_SUMMARY");
      insertObject(sequence, entityID, 0, "WAFER_EVENT_EID", waferID);
      insertObject(sequence, entityID, 0, "TOUCHDOWN_COUNT", eventGroup);
      Long tmp = (Long) fields.get("PART_CNT");
      if( tmp != null) if( tmp != 4294967295L)insertObject(sequence, entityID, 0, "PART_COUNT", fields.get("PART_CNT"));
      tmp = (Long) fields.get("RTST_CNT");
      if( tmp != null) if( tmp != 4294967295L)insertObject(sequence, entityID, 0, "RETEST_PART_COUNT", fields.get("RTST_CNT"));
      tmp = (Long) fields.get("ABRT_CNT");
      if( tmp != null) if( tmp != 4294967295L)insertObject(sequence, entityID, 0, "ABORT_PART_COUNT", fields.get("ABRT_CNT"));
      tmp = (Long) fields.get("GOOD_CNT");
      if( tmp != null) if( tmp != 4294967295L)insertObject(sequence, entityID, 0, "GOOD_PART_COUNT", fields.get("GOOD_CNT"));
      tmp = (Long) fields.get("FUNC_CNT");
      //if( tmp != null) if( tmp != 4294967295L)insertObject(sequence, entityID, 0, "FUNCTIONAL_PART_COUNT", fields.get("FUNC_CNT"));
      waferID = 0; // reset for next wafer
    }
    
    /**  Handles WCR Record mapping
     * 
     */
    private void insertWcr(HashMap  fields) {
      entityID++;
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "WAFER_INFO");
      insertObject(sequence, entityID, 0, "WAFER_SIZE", fields.get("WAFR_SIZ"));
      insertObject(sequence, entityID, 0, "DIE_HEIGHT", fields.get("DIE_HT"));
      insertObject(sequence, entityID, 0, "DIE_WIDTH" , fields.get("DIE_WID"));
      Long time = (Long) fields.get("WF_UNITS");
      if(time.equals(0L)) insertObject(sequence, entityID, 0, "WAFER_UNITS", "NONE");
      if(time.equals(1L)) insertObject(sequence, entityID, 0, "WAFER_UNITS", "in");
      if(time.equals(2L)) insertObject(sequence, entityID, 0, "WAFER_UNITS", "cm");
      if(time.equals(3L)) insertObject(sequence, entityID, 0, "WAFER_UNITS", "mm");
      if(time.equals(4L)) insertObject(sequence, entityID, 0, "WAFER_UNITS", "mil");
      insertObject(sequence, entityID, 0, "WAFER_FLAT", fields.get("WF_FLAT"));
      insertObject(sequence, entityID, 0, "CENTER_X", fields.get("CENTER_X"));
      insertObject(sequence, entityID, 0, "CENTER_Y", fields.get("CENTER_Y"));
      insertObject(sequence, entityID, 0, "POSITIVE_X_DIRECTION", fields.get("POS_X"));
      insertObject(sequence, entityID, 0, "POSITIVE_Y_DIRECTION", fields.get("POS_Y"));

    }
    
    //now lets go for the test synopsis record
    private void insertTsr(HashMap  fields) {
      // first get the test number as this is the reference
      Integer tmp = ((Long) fields.get("TEST_NUM")).intValue();
      Integer testNumber = getTestEntityFor(tmp,(String) fields.get("TEST_NAM"), -1L);
      entityID++;  // get the next entity ID and do the test synopsis
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_SUMMARY");
      insertObject(sequence, entityID, 0, "TEST_INFO_EID", testNumber);
      int head  = ((Long) fields.get("HEAD_NUM")).intValue();
      if(head == 255){
        insertObject(sequence, entityID, 0, "SITE", "all");       
      }else {
        int site = ((Long) fields.get("SITE_NUM")).intValue();
        insertObject(sequence, entityID, 0, "SITE", site);
      }
      String tName = ((String) fields.get("TEST_NAM"));
      if(_testByNumber && !testName.containsKey(testNumber)) {
        insertObject(sequence, testNumber, 0, "TEST_NAME", tName); 
        testName.put(testNumber, true);
      }
      int opt;
      if(fields.get("OPT_FLAG") == null) {
        opt = 0;
      }else{
        opt =  ((Byte) fields.get("OPT_FLAG"))  & 0x37;
      }
      //insertObject(sequence, entityID, 0, "OPT_FLAG", opt);
      Iterator<String> iterFieldNames = fields.keySet().iterator();
      insertObject(sequence, entityID, 0, "TSR_TEST_NAME", tName);
      while(iterFieldNames.hasNext()) {
        String newName = null;
        Object fieldValue = null;
        String fieldName = (String)iterFieldNames.next();
        fieldValue = fields.get(fieldName); // set to null to drop
        // start the name mapping section 
        if(fieldName.equals("TEST_MAX")) {
          if( ((opt & 0x02) == 0)) {
            newName = "MAXIMUM_TEST_VALUE";
            insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("TEST_MIN")) {
          if( ((opt & 0x01) == 0)) {
            newName = "MINIMUM_TEST_VALUE";
            insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("FAIL_CNT")) {
          newName = "TEST_FAIL_COUNT";
          if( ((Number)fieldValue).longValue() != 4294967295L) insertObject(sequence, entityID, 0, newName, fieldValue);
        }else if(fieldName.equals("ALRM_CNT")) {
        newName = "TEST_ALARM_COUNT";
          if( ((Number)fieldValue).longValue() != 4294967295L) {
             insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("EXEC_CNT")) {
          newName = "TEST_EXECUTION_COUNT";
          if( ((Number)fieldValue).longValue() != 4294967295L) insertObject(sequence, entityID, 0, newName, fieldValue);
        }else if(fieldName.equals("TST_SUMS")) {
          if( ((opt & 0x10) == 0)) {
            newName = "SUM_TEST_VALUES";
            insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("TST_SQRS")) {
          if( ((opt & 0x20) == 0)) {
            newName = "SUMOFSQUARES_TEST_VALUES";
            insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("TEST_TIM")) {
          newName = "AVERAGE_TEST_TIME";
          if( ((opt & 0x04) == 0)) {
            insertObject(sequence, entityID, 0, newName, fieldValue);
          }
        }else if(fieldName.equals("TEST_NUM")) {
          newName = "TEST_ID";
          insertObject(sequence, entityID, 0, newName, fieldValue);
          }else if(fieldName.equals("SEQ_NAME")) {
          newName = "TEST_SEQUENCER_NAME";
          insertObject(sequence, entityID, 0, newName, fieldValue);
        }else if(fieldName.equals("TEST_LBL")) {
          newName = "TEST_LABEL";
          insertObject(sequence, entityID, 0, newName, fieldValue);
        }
      }
    }
    private void insertPir(HashMap  fields) {
      entityID++;// start a new testEvent
      int site = ((Long) fields.get("SITE_NUM")).intValue();
      currentTE[site & 0xff] = entityID;
      if(!inGroup) {
        inGroup = true;
        eventGroup++;
      }
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_EVENT");
      insertObject(sequence, entityID, 0, "SITE", site);
      insertObject(sequence, entityID, 0, "EVENT_GROUP", eventGroup);
      if( waferID != 0) {
        insertObject(sequence, entityID, 0, "WAFER_EVENT_EID", waferID);        
      }
    }
    
    private void insertPrr(HashMap  fields) {
      if(inGroup) {
        inGroup = false;
        if((fields.get("TEST_T") != null) && ((Long)fields.get("TEST_T") != 0L)) {          
          Long timeSecs = (Long)fields.get("TEST_T") / 100L; // millis to 10us
            realSequence = (timeSecs) + realSequence;
          }
      }

      int site = ((Long) fields.get("SITE_NUM")).intValue();
      int te = currentTE[site & 0xff];
      insertObject(sequence, te, 0, "TEST_LIMITS_EID", currentLimits);
      insertObject(sequence, te, 0, "HARD_BIN", fields.get("HARD_BIN"));
      int tmp = ((Integer) fields.get("X_COORD"));
      if(tmp != -32768) {
        insertObject(sequence, te, 0, "X", tmp);
      }
      tmp = ((Integer) fields.get("Y_COORD"));
      if(tmp != -32768) {
        insertObject(sequence, te, 0, "Y", tmp);
      }  
      long tmpL = ((Long) fields.get("SOFT_BIN"));
      if(tmpL != 65535L) {
        insertObject(sequence, te, 0, "SOFT_BIN", tmpL);
      }  
      insertObject(sequence, te, 0, "UNIT_TEST_TIME", fields.get("TEST_T"));
      insertObject(sequence, te, 0, "NUM_TESTS", fields.get("NUM_TEST"));
      insertObject(sequence, te, 0, "PART_ID", fields.get("PART_ID"));
      insertObject(sequence, te, 0, "PART_TEXT", fields.get("PART_TXT"));
      
      int flags = ((Byte) fields.get("PART_FLG")).intValue();
      if((flags & 0x18) == 0x08) {
        insertObject(sequence, te, 0, "PF", "FAIL", "F");
      }
      if((flags & 0x18) == 0x0) {
        insertObject(sequence, te, 0, "PF", "PASS", "P");
      }
      if((flags & 0x14) == 0x10) {
        insertObject(sequence, te, 0, "PF", "TESTED", "T");
      }
      if((flags & 0x04) == 0x04) {
        insertObject(sequence, te, 0, "PF", "ABORT", "A");
      }
      if((flags & 0x03) != 0x0) {
        insertObject(sequence, te, 0, "RETEST_CODE", "RETEST", "R");
      }
    }
    
    private void insertPtr(HashMap  fields) {
      Long site = (Long) fields.get("SITE_NUM");
      int te = currentTE[site.intValue()];
      Integer tmp1 = ((Long) fields.get("TEST_NUM")).intValue();
      Integer testNumber = getTestEntityFor(tmp1,(String) fields.get("TEST_TXT"), site);
      Byte parm  = ((Byte) fields.get("PARM_FLG")); // can use for high low fail
      int flags = ((Byte) fields.get("TEST_FLG")).intValue();
      if((flags & 0x2d) != 0) {
        //insertObject(sequence, te, testNumber, "CALC_TEST_FLAG", flags);
      }
      String pf = null;
      if((flags & 0x40) == 0){// pass/fail flag is valid
        if((flags & 0x80) == 0x80) {
          pf = "F";
        }else{
          pf = "P" ; 
        }
      } 
      Double result = null;  // check for valid result
      if((flags & 0x3f) == 0) result = ((Number) fields.get("RESULT")).doubleValue();
      
      if((flags & 0x10) == 0) { // test performed
        insertObject(sequence, te, testNumber, "R", result, pf);
      }
           
 
      insertObject(sequence, te, testNumber, "ALARM_ID", fields.get("ALARM_ID"));

      if(!testUpdate.containsKey(testNumber)) {
        testUpdate.put(testNumber, 0L);
          //insertObject(sequence, testNumber, 0, "TEST_NUMBER", fields.get("TEST_NUM"));
          insertObject(sequence, testNumber, 0, "TEST_TYPE", "PTR");
          insertObject(sequence, testNumber, 0, "TEST_RESULT_TYPE", "CONTINIOUS");
        // insertObject(sequence, testNumber, 0, "R_PRNTFMT", fields.get("C_RESFMT"));
        //insertObject(sequence, testNumber, 0, "LL_PRNTFMT", fields.get("C_LLMFMT"));
        // insertObject(sequence, testNumber, 0,"UL_PRNTFMT", fields.get("C_HLMFMT"));
          insertObject(sequence,  testNumber,0 , "TEST_UNITS", fields.get("UNITS"));
          insertObject(sequence,  testNumber,0 , "TEST_UNITS_LABEL", fields.get("UNITS"));
        Byte opt = (Byte) fields.get("OPT_FLAG");
        if( opt != null) {
          if(currentLimits == 0) {// add a limits entity
            entityID++;
            currentLimits = entityID;
            insertObject(sequence, currentLimits, 0, "ENTITY_TYPE", "TEST_LIMITS");
            insertObject(sequence, currentLimits, 0, "LIMITS_NAME", "standard limits");
          }
          if( (opt & 0x80) == 0) { // insert high limit
            insertObject(sequence, currentLimits, testNumber, "UL", fields.get("HI_LIMIT"));  
            //insertObject(sequence, testNumber, 0, "UL_SCALE", fields.get("HLM_SCAL"));
          }
          if( (opt & 0x40) == 0) { // insert low  limit
            insertObject(sequence, currentLimits , testNumber, "LL", fields.get("LO_LIMIT"));
            //insertObject(sequence, testNumber, 0, "LL_SCALE", fields.get("LLM_SCAL"));
          }
          if( (opt & 0x01) == 0) { // insert results scale
            insertObject(sequence, testNumber, 0, "R_SCALE", fields.get("RES_SCAL"));
          }
          if( (opt & 0x04) == 0) { // insert lo spec
            insertObject(sequence, testNumber, 0, "LSL", fields.get("LO_SPEC"));
          }
          if( (opt & 0x08) == 0) { // insert hi spec
            insertObject(sequence, testNumber, 0, "USL", fields.get("HI_SPEC"));
          }
        }
      }
      // now update the testText
      updateTextText(testNumber,site.intValue(), (String)fields.get("TEST_TXT"));
      /*if(testUpdate.containsKey(testNumber)) {
        Long test = testUpdate.get(testNumber);
        if((((int) (test >> site)) & 0x1) == 0){  // check to see if the flag is set "
          insertObject(sequence, testNumber, site.intValue(), "TEST_TEXT", fields.get("TEST_TXT"));
          test = testUpdate.get(testNumber);
          testUpdate.put(testNumber, (1L << site) + test);
        }
      }*/
    }
    
    private void insertPcr(HashMap  fields) {
      entityID++;// start a new entity
      Long head  = (Long) fields.get("HEAD_NUM");
      Long tmp ;
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "RUN_SUMMARY");
      insertObject(sequence, entityID, 0, "PART_COUNT", fields.get("PART_CNT"));
      insertObject(sequence, runID, 0, "PART_COUNT", fields.get("PART_CNT"));
      tmp = (Long) fields.get("RTST_CNT");
      if( tmp != null) if( tmp != 4294967295L) insertObject(sequence, entityID, 0, "RETEST_PART_COUNT", fields.get("RTST_CNT"));
      tmp = (Long) fields.get("ABRT_CNT");
      if( tmp != null) if(  tmp != 4294967295L)insertObject(sequence, entityID, 0, "ABORT_PART_COUNT", fields.get("ABRT_CNT"));
      tmp = (Long) fields.get("GOOD_CNT");
      if( tmp != null){
        if( tmp != 4294967295L){
          insertObject(sequence, entityID, 0, "GOOD_PART_COUNT", fields.get("GOOD_CNT"));
          Long total = (Long) fields.get("PART_CNT");
          if(total > 0){
	          Long yield = (tmp * 100L) / total;
	          insertObject(sequence, runID, 0, "YIELD", yield.intValue());
          }
        }
      }
      if(head != 255L) {
        insertObject(sequence, entityID, 0, "SITE", fields.get("SITE_NUM"));
      }
      tmp = (Long) fields.get("FUNC_CNT");
      if( tmp != null) if(tmp != 4294967295L)insertObject(sequence, entityID, 0, "FUNCTIONAL_PART_COUNT", fields.get("FUNC_CNT")); 
    }
    
    private void insertAtr(HashMap  fields) {
      entityID++;// start a new entity
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "AUDIT_INFO");
      long time = (Long) fields.get("MOD_TIM");
      if(time != 0) {
        insertObject(time, entityID, 0, "DATE_DTTM_UTC", convertUnixTimeStampGMT(time));
        insertObject(time, entityID, 0, "DATE_DTTM", convertUnixTimeStamp(time));
      }
      insertObject(sequence, entityID, 0, "CMD", fields.get("CMD_LINE"));
    }
    
    private void insertBps(HashMap  fields) {
      if(bpsID == 0){
        entityID++;
        bpsID = entityID;
        insertObject(sequence, bpsID, 0, "ENTITY_TYPE", "BPS_RECS");
      }
      insertObject(sequence, bpsID, 0, "SEQ_NAME", fields.get("SEQ_NAME"));
    }
    
    private void insertEps(HashMap  fields) {
      if(epsID == 0){
        entityID++;
        epsID = entityID;
        insertObject(sequence, epsID, 0, "ENTITY_TYPE", "DTR_RECS");
      }
      insertObject(sequence, epsID, 0, "EPS", "MARKER");
    }
    
    
    private void insertDtr(HashMap  fields) {
      if(dtrID == 0){
        entityID++;
        dtrID = entityID;
        insertObject(sequence, dtrID, 0, "ENTITY_TYPE", "DTR_RECS");
      }
      insertObject(sequence, dtrID, 0, "DTR_TXT", fields.get("TEXT_DAT"));
    }
    
    private void insertGdr(HashMap  fields) {
      entityID++;// start a new entity
      Object[] tmp = (Object[]) fields.get("GEN_DATA");
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "GDR_DATA");
      insertObject(sequence, entityID, 0, "FIELD_CNT", fields.get("FLD_CNT"));
      Integer max = tmp.length;
      for(int i = 0; i < max; i++) {
        insertObject(sequence, entityID, i + 1, "DATA_ITEM", tmp[i]);    
      }
    }
    
    private void insertSdr(HashMap  fields) {
      Iterator<String> iterFieldNames = fields.keySet().iterator();
      entityID++;  // get the next entity ID and do the test synopsis
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "SITE_INFO");
      insertObject(sequence, entityID, 0, "SITE_GROUP_NUMBER", fields.get("SITE_GRP"));
      insertObject(sequence, entityID, 0, "SITE_COUNT", fields.get("SITE_CNT"));
      insertObject(sequence, runID, 0, "HANDLER_TYPE", fields.get("HAND_TYP"));
      insertObject(sequence, runID, 0, "HANDLER_ID", fields.get("HAND_ID"));
      insertObject(sequence, runID, 0, "CARD_TYPE", fields.get("CARD_TYP"));
      insertObject(sequence, runID, 0, "CARD_ID", fields.get("CARD_ID"));
      insertObject(sequence, runID, 0, "LOADBOARD_TYPE", fields.get("LOAD_TYP"));
      insertObject(sequence, runID, 0, "LOADBOARD_ID", fields.get("LOAD_ID"));
      insertObject(sequence, runID, 0, "DIB_TYPE", fields.get("DIB_TYP"));
      insertObject(sequence, runID, 0, "DIB_ID", fields.get("DIB_ID"));
      insertObject(sequence, runID, 0, "INTERFACE_CABLE_TYPE", fields.get("CABL_TYP"));
      insertObject(sequence, runID, 0, "INTERFACE_CABLE_ID", fields.get("CABL_ID"));
      insertObject(sequence, runID, 0, "CONTACTOR_TYPE", fields.get("CONT_TYP"));
      insertObject(sequence, runID, 0, "CONTACTOR_ID", fields.get("CONT_ID"));
      insertObject(sequence, runID, 0, "LASER_TYPE", fields.get("LASR_TYP"));
      insertObject(sequence, runID, 0, "LASER_ID", fields.get("LASR_ID"));
      //insertObject(sequence, entityID, 0, "EXTR_ID", fields.get("EXTR_ID"));
      //insertObject(sequence, entityID, 0, "EXTR_TYP", fields.get("EXTR_TYP"));
// TODO SITE_NUM which is an array
      int count = ((Number) fields.get("SITE_CNT")).intValue();
      if( count > 0) {
        Object[] list = (Object[]) fields.get("SITE_NUM");
        for( int i = 0; i < count; i++) {
          insertObject(sequence, entityID , (i + 1), "ACTIVE_SITE", list[i]);  // first test is the common info
        }
      }
    }
    
    private void insertHbr(HashMap  fields) {
      entityID++;// start a new entity
      Long head  = (Long) fields.get("HEAD_NUM");
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "HARD_BIN");
      insertObject(sequence, entityID, 0, "BIN_NUMBER", fields.get("HBIN_NUM"));
      insertObject(sequence, entityID, 0, "BIN_COUNT", fields.get("HBIN_CNT"));
      insertObject(sequence, entityID, 0, "BIN_NAME", fields.get("HBIN_NAM"));
      insertObject(sequence, entityID, 0, "BIN_PF", fields.get("HBIN_PF"));
      if(head != 255L) {
        insertObject(sequence, entityID, 0, "SITE", fields.get("SITE_NUM"));
      }
    }
    
    private void insertRdr(HashMap  fields) {
      entityID++;// start a new entity
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "RESCREEN");
      Number tmpStr = (Number) fields.get("NUM_BINS");
      //insertObject(sequence, entityID , 0, "RESCREEN_BIN_COUNT", tmpStr);
      int count = tmpStr.intValue();
      if( count > 0) {
        Object[] list = (Object[]) fields.get("RTST_BIN");
        for( int i = 0; i < count; i++) {
          insertObject(sequence, entityID , (i + 1), "RESCREEN_BIN", list[i]);
        }
      }
    }
    
    private void insertSbr(HashMap  fields) {
      entityID++;// start a new entity
      Long head  = (Long) fields.get("HEAD_NUM");
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "SOFT_BIN");
      insertObject(sequence, entityID, 0, "BIN_NUMBER", fields.get("SBIN_NUM"));
      insertObject(sequence, entityID, 0, "BIN_COUNT", fields.get("SBIN_CNT"));
      insertObject(sequence, entityID, 0, "BIN_NAME", fields.get("SBIN_NAM"));
      insertObject(sequence, entityID, 0, "BIN_PF", fields.get("SBIN_PF"));
      if(head != 255L) {
        insertObject(sequence, entityID, 0, "SITE", fields.get("SITE_NUM"));
      }

    }
    private void insertPlr(HashMap  fields) {
      entityID++;// start a new entity
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "PIN_LIST");     
      Integer max = ((Number) fields.get("GRP_CNT")).intValue();
      Object [] pmgi =  (Object[]) fields.get("GRP_INDX");
      Object [] pmgm =  (Object[]) fields.get("GRP_MODE");
      Object [] pmgr =  (Object[]) fields.get("GRP_RADX");
      Object [] pmgpc =  (Object[]) fields.get("PGM_CHAR");
      Object [] pmgrc =  (Object[]) fields.get("RTN_CHAR");
      Object [] pmgpl =  (Object[]) fields.get("PGM_CHAL");
      Object [] pmgrl =  (Object[]) fields.get("RTN_CHAL");
   
      for(int i = 0; i < max; i++) {
        if((Long) pmgi[i] != 0L) insertObject(sequence, entityID, i + 1, "GROUP_INDEX", pmgi[i]);
        if((Long) pmgm[i] != 0L) insertObject(sequence, entityID, i + 1, "GROUP_MODE", pmgm[i]);
        if((Long) pmgr[i] != 0L) insertObject(sequence, entityID, i + 1, "GROUP_RADIX", pmgr[i]);
        if(pmgpc != null) insertObject(sequence, entityID, i + 1, "PROGRAM_CHAR", pmgpc[i]);
        if(pmgrc != null) insertObject(sequence, entityID, i + 1, "RETURN_CHAR", pmgrc[i]);
        if(pmgpl != null) insertObject(sequence, entityID, i + 1, "PROGRAM_CHAL", pmgpl[i]);
        if(pmgrl != null) insertObject(sequence, entityID, i + 1, "RETURN_CHAL", pmgrl[i]);
      }
    }
    private void insertPgr(HashMap  fields) {
      entityID++;// start a new entity
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "PIN_GROUP");
      insertObject(sequence, entityID, 0, "GROUP_NAME", fields.get("GRP_NAM"));
      insertObject(sequence, entityID, 0, "GROUP_INDEX", fields.get("GRP_INDX"));      
      Integer max = ((Number) fields.get("INDX_CNT")).intValue();
      Object [] pmrs =  (Object[]) fields.get("PMR_INDX");
      for(int i = 0; i < max; i++) {
        insertObject(sequence, entityID, i + 1, "PMR_INDEX", pmrs[i]);    
      }
    }
    private void insertPmr(HashMap  fields) {
      entityID++;// start a new entity
      Number  site = (Number) fields.get("SITE_NUM");
      if( site == null) {
        site = 1;
      }
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "PIN_MAP");
      insertObject(sequence, entityID, 0, "SITE", site.intValue());
      insertObject(sequence, entityID, 0, "CHANNEL_NAME", fields.get("CHAN_NAM")); 
      insertObject(sequence, entityID, 0, "PHYSICAL_NAME", fields.get("PHY_NAM")); 
      insertObject(sequence, entityID, 0, "PMR_INDEX", fields.get("PMR_INDX"));
      insertObject(sequence, entityID, 0, "CHANNEL_TYPE", fields.get("CHAN_TYP"));
      insertObject(sequence, entityID, 0, "LOGIGAL_NAME", fields.get("LOG_NAM"));
    }
    
    private void insertMpr(HashMap  fields) {
      Long site = (Long) fields.get("SITE_NUM");
      int te = currentTE[site.intValue()];
      Integer tmp1 = ((Long) fields.get("TEST_NUM")).intValue();
      Integer testNumber = getTestEntityFor(tmp1,(String) fields.get("TEST_TXT"), site);
      insertObject(sequence, testNumber, 0, "TEST_RESULT_TYPE", "CONTINIOUS");
      Byte parm  = ((Byte) fields.get("PARM_FLG")); // can use for high low fail
      int flags = ((Byte) fields.get("TEST_FLG")).intValue();
      String pf = "P";
      if((flags & 0xC0) == 0x80) {
        insertObject(sequence, te, testNumber, "CALC_TEST_FLAG", flags & 0xff);
        pf = "F";
      }
      if((flags & 0x20) == 0x20) pf = "A";
      if((flags & 0x50) == 0x50) pf = "S";
      if((flags & 0x7a) == 0x40) pf = "T";
      if((flags & 0xa0) == 0xa0) pf =  "FA";
      
      String tmpStr = (String) fields.get("ALARM_ID");
      if(tmpStr != null) { 
          if(!(tmpStr.trim().isEmpty())) {
              insertObject(sequence, te, testNumber, "ALARM", tmpStr);
          }
      }
      int count = ((Number) fields.get("RSLT_CNT")).intValue();
      int statcount = ((Number) fields.get("RTN_ICNT")).intValue();
      String flag = "P";
      if(count != 1)insertObject(sequence, te, testNumber, "R", null, pf);// overall result flag
      if(statcount == count && count > 0) {
        Object[] list = (Object[]) fields.get("RTN_RSLT");
        Object[] state = (Object[]) fields.get("RTN_STAT");
        if(count == 1){
          Integer bin = (Integer)state[0];
          int binVal = (bin == null) ? 4 : bin.intValue(); 
          switch(binVal) {
            case 0:  flag = "PL"; break;         
            case 1:  flag = "PH"; break;         
            case 2:  flag = "PM"; break;  
            case 3:  flag = "PG"; break;        
            case 4:  flag = "TX"; break;         
            case 5:  flag = "FL"; break;        
            case 6:  flag = "FH"; break;  
            case 7:  flag = "FM"; break;        
            case 8:  flag = "FG"; break;
            case 9:  flag = "TO"; break;  
            case 10: flag = "TS"; break;   
          }
          insertObject(sequence, te , testNumber, "R", list[0], pf);
        }else{
          for( int i = 0; i < count; i++) {
            Integer bin = (Integer)state[i];
            int binVal = (bin == null) ? 4 : bin.intValue(); 
            switch(binVal) {
              case 0:  flag = "PL"; break;         
              case 1:  flag = "PH"; break;         
              case 2:  flag = "PM"; break;  
              case 3:  flag = "PG"; break;        
              case 4:  flag = "TX"; break;         
              case 5:  flag = "FL"; break;        
              case 6:  flag = "FH"; break;  
              case 7:  flag = "FM"; break;        
              case 8:  flag = "FG"; break;
              case 9:  flag = "TO"; break;  
              case 10: flag = "TS"; break;   
            }
            insertObject(sequence, te , (testNumber + i + 1), "R", list[i], flag);  // first test is the common info
          }
        }
      }else{
        if( count > 0) {
          Object[] list = (Object[]) fields.get("RTN_RSLT");
          if ( count == 1){
            insertObject(sequence, te , testNumber, "R", list[0]);
          }else{
            for( int i = 0; i < count; i++) {
              insertObject(sequence, te , (testNumber + i + 1), "R", list[i]);  // first test is the common info
            }
          }
        }
        if( statcount > 0) {
          Object[] list = (Object[]) fields.get("RTN_STAT");
          if(count == 1){
            insertObject(sequence, te , testNumber, "TEST_FLAG", list[0]);
          }else{
            for( int i = 0; i < statcount; i++) {
              insertObject(sequence, te , (testNumber + i + 1), "TEST_FLAG", list[i]);  // first test is the common info
            }
          }
        }
      }
      if(!testUpdate.containsKey(testNumber)) {
        insertObject(sequence, testNumber, 0, "TEST_TYPE", "MPR");
        //insertObject(sequence, testNumber, 0,"TEST_NUMBER", fields.get("TEST_NUM"));
        testUpdate.put(testNumber, 0L);
        count = ((Number) fields.get("RSLT_CNT")).intValue();
        insertObject(sequence, testNumber , 0, "RESULT_COUNT", count);
        insertObject(sequence, testNumber , 0, "RTN_ICNT", fields.get("RTN_ICNT"));
        //insertObject(sequence, testNumber, 0, "TEST_TEXT", fields.get("TEST_TXT"));
        Object[] list = (Object[]) fields.get("RTN_INDX");
        Byte opt = (Byte) fields.get("OPT_FLAG");
        if( opt != null) {
          if( (opt & 0x01) == 0) { // insert smhoo info
            insertObject(sequence, testNumber, 0, "START_IN", fields.get("START_IN")); 
            insertObject(sequence, testNumber, 0, "INCR_IN", fields.get("INCR_IN"));
          }
        }
        int subtest = 0;
        for( int j = 0 ; j < count ; j++) {
          if(count == 1){
            subtest = testNumber;
          }else {
        	if(j == 0){
                subtest = testNumber;
        	}else{
                entityID++;// start a new entity
                subtest = entityID;	
        	}
            insertObject(sequence, subtest, 0,  "ENTITY_TYPE", "TEST_INFO");
            insertObject(sequence, subtest, 0,  "TYPE", "SUBTEST");
            insertObject(sequence, subtest, 0,  "TEST_EID", testNumber);
          }
          insertObject(sequence, subtest, 0,  "RTN_INDX", list[j]);
          insertObject(sequence, subtest, 0,  "TEST_ORDER", j);
          //insertObject(sequence, subtest, 0,  "R_PRNTFMT", fields.get("C_RESFMT"));
          //insertObject(sequence, subtest, 0,  "LL_PRNTFMT", fields.get("C_LLMFMT"));
          //insertObject(sequence, subtest, 0,  "UL_PRNTFMT", fields.get("C_HLMFMT"));
          insertObject(sequence, subtest, 0,  "UNITS", fields.get("UNITS"));
          insertObject(sequence, subtest, 0,  "UNITS_IN", fields.get("UNITS_IN"));
          if( opt != null) {
            if(currentLimits == 0) {// add a limits entity
              entityID++;
              currentLimits = entityID;
              insertObject(sequence, currentLimits, 0, "ENTITY_TYPE", "TEST_LIMITS");
              insertObject(sequence, currentLimits, 0, "LIMITS_NAME", "standard limits");
            }
            if( (opt & 0xa0) == 0) { // insert high limit
              insertObject(sequence, currentLimits, testNumber, "UL", fields.get("HI_LIMIT")); 
              //insertObject(sequence, subtest, 0, "UL", fields.get("HI_LIMIT")); 
              //insertObject(sequence, subtest, 0, "UL_SCALE", fields.get("HLM_SCAL"));
            }
            if( (opt & 0x50) == 0) { // insert low  limit
              insertObject(sequence, currentLimits , testNumber, "LL", fields.get("LO_LIMIT"));
              //insertObject(sequence, subtest, 0, "LL", fields.get("LO_LIMIT"));
              //insertObject(sequence, subtest, 0, "LL_SCALE", fields.get("LLM_SCAL"));
            }
            if( (opt & 0x01) == 0) { // insert results scale
              insertObject(sequence, subtest, 0, "R_SCALE", fields.get("RES_SCAL"));
            }
            if( (opt & 0x04) == 0) { // insert lo spec
              insertObject(sequence, subtest, 0, "LSL", fields.get("LO_SPEC"));
            }
            if( (opt & 0x08) == 0) { // insert hi spec
              insertObject(sequence, subtest, 0, "HSL", fields.get("HI_SPEC"));
            }
          }
        }
      }
      // now update the testText
      updateTextText(testNumber,site.intValue(), (String)fields.get("TEST_TXT"));
     /* if(testUpdate.containsKey(testNumber)) {
        Long test = testUpdate.get(testNumber);
        if((((int) (test >> site)) & 0x1) == 0){  // check to see if the flag is set "
          insertObject(sequence, testNumber, site.intValue(), "TEST_TEXT", fields.get("TEST_TXT"));
          test = testUpdate.get(testNumber);
          testUpdate.put(testNumber, (1L << site) + test);
        }
      } */
    }
    private void insertFtr(HashMap  fields) {
      Long site = (Long) fields.get("SITE_NUM");
      int te = currentTE[site.intValue()];
      Integer tmp1 = ((Long) fields.get("TEST_NUM")).intValue();
      Integer testNumber = getTestEntityFor(tmp1,(String) fields.get("TEST_TXT"), site);
      insertObject(sequence, testNumber, 0, "TEST_RESULT_TYPE", "FUNCTIONAL"); 
      insertObject(sequence, te, testNumber, "FAIL_PIN", fields.get("FAIL_PIN"));
      insertObject(sequence, te, testNumber, "VECT_NAM", fields.get("VECT_NAM"));
      insertObject(sequence, te, testNumber, "TIME_SET", fields.get("TIME_SET"));
      insertObject(sequence, te, testNumber, "OP_CODE", fields.get("OP_CODE"));
      insertObject(sequence, te, testNumber, "PROG_TXT", fields.get("PROG_TXT"));
      insertObject(sequence, te, testNumber, "RSLT_TXT", fields.get("RSLT_TXT"));
      int tmp = ((Long)fields.get("RTN_ICNT")).intValue();
      if(tmp != 0) {
        insertObject(sequence, te, testNumber, "RTN_ICNT", fields.get("RTN_ICNT"));
        insertObject(sequence, te, testNumber, "RTN_INDX", fields.get("RTN_INDX")); 
        insertObject(sequence, te, testNumber, "RTN_STAT", fields.get("RTN_STAT"));
      }
      tmp = ((Long) fields.get("PGM_ICNT")).intValue();
      if(tmp != 0) {
        insertObject(sequence, te, testNumber, "PGM_ICNT", fields.get("PGM_ICNT"));
        insertObject(sequence, te, testNumber, "PGM_INDX", fields.get("PGM_INDX")); 
        insertObject(sequence, te, testNumber, "PGM_STAT", fields.get("PGM_STAT"));
      }
      //TODO discard if 255
      insertObject(sequence, te, testNumber, "PATG_NUM", fields.get("PATG_NUM"));
      // these are arrays
 
      insertObject(sequence, te, testNumber, "SPIN_MAP", fields.get("SPIN_MAP")); 
      
      int opt  = ((Byte) fields.get("OPT_FLAG")); 
      if((opt & 0x01) == 0) {
        insertObject(sequence, te, testNumber, "CYCL_CNT", fields.get("CYCL_CNT"));
      }
      if((opt & 0x02) == 0) {
        insertObject(sequence, te, testNumber, "REL_VADR", fields.get("REL_VADR"));
      }
      if((opt & 0x04) == 0) {
        insertObject(sequence, te, testNumber, "REPT_CNT", fields.get("REPT_CNT"));
      }
      if((opt & 0x08) == 0) {
        insertObject(sequence, te, testNumber, "NUM_FAIL", fields.get("NUM_FAIL"));
      }
      if((opt & 0x10) == 0) {
        insertObject(sequence, te, testNumber, "XFAIL_AD", fields.get("XFAIL_AD"));
      }
      if((opt & 0x20) == 0) {
        insertObject(sequence, te, testNumber, "YFAIL_AD", fields.get("YFAIL_AD"));
      }
      if((opt & 0x40) == 0) {
        insertObject(sequence, te, testNumber, "VECT_OFF", fields.get("VECT_OFF"));
      }

      int flags = ((Byte) fields.get("TEST_FLG")).intValue();
      if((flags & 0x2d) != 0) {
        insertObject(sequence, te, testNumber, "CALC_TEST_FLAG", flags);
      }
      String pf = null;
      if((flags & 0x40) == 0){// pass/fail flag is valid
        if((flags & 0x80) == 0x80) {
          pf = "F";
        }else{
          pf = "P" ; 
        }
      } 
      if((flags & 0x50) == 0) {
        insertObject(sequence, te, testNumber, "R", null, pf);
      }
      if((flags & 0x20) == 0x20) {
        insertObject(sequence, te, testNumber, "R", null, "A");
      }
      if((flags & 0x50) == 0x50) {
        insertObject(sequence, te, testNumber, "R", null, "S");
      }
      if((flags & 0x7a) == 0x40) {
        insertObject(sequence, te, testNumber, "R", null, "T");
      }
      if((flags & 0xa0) == 0xa0) {
        insertObject(sequence, te, testNumber, "R", null, "FA");
      }
      String tmpStr = (String) fields.get("ALARM_ID");
      if(tmpStr != null) { 
          if(!(tmpStr.trim().isEmpty())) {
              insertObject(sequence, te, testNumber, "ALARM", tmpStr);
          }
      }
      if(!testUpdate.containsKey(testNumber)) {
        //insertObject(sequence, testNumber, 0, "TEST_NUMBER", fields.get("TEST_NUM"));
        insertObject(sequence, testNumber, 0, "TEST_TYPE", "FTR");
        testUpdate.put(testNumber, 0L);
      }
      // now update the testText
      updateTextText(testNumber,site.intValue(), (String)fields.get("TEST_TXT"));
    }
}
