package ri.stdf;

import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import ri.util.Emit;

/**A mapping between STDF4 and STDF5 that is initialized from an XML file.
 * This reads through the XML file and builds a collection of mappings that 
 * define the operations needed.
 */
public class StdfMapping extends DefaultHandler {

  private HashMap<String, HashMap<String, String>> _map = new HashMap<String, HashMap<String, String>>(); //mappings of stdf4 to stdf5 assignments
  private Stack<String> tagStack=new Stack<String>(); //the current tag stack
  private StringBuffer _sbValue = new StringBuffer(); //temp buffer for the value parameter
  private HashMap<String, String> _currentMap=null;
  
  
  /**Constr*/
  public StdfMapping() {
    InputStream stream=null;
    try { 
      stream=this.getClass().getResourceAsStream("stdf4to5Mapping.xml"); //XML mapping for STDF4 to STDF5
      InputSource source=new InputSource(stream);
      SAXParser parser=SAXParserFactory.newInstance().newSAXParser();
      parser.parse(source, this);
    }
    catch(Exception e) {
      Emit.out("Error reading the STDF4 to StDF5 mapping file: "+e);
      e.printStackTrace(); 
    }
  }

  
  /**Given an STDF4 type (e.g. "Mir" "Wir" "Mrr" "Wrr" "Wcr" "Tsr" "Pir" "Prr" "Ptr" "Pcr" 
   * "Atr" "Bps" "Eps" "Dtr" "Gdr" "Sdr" "Hbr" "Rdr" "Sbr" "Pgr" "Pmr")
   * returns the mapping associated with that type*/
  public HashMap<String, String> getMapFor(String stdf4Type) {
    return _map.get(stdf4Type.toLowerCase());
  }
  
  /**XML Parsing. Called at the start of an element*/
  @Override
  public void startElement(String uri, String localName, String qName, Attributes attributes) {
  tagStack.push(qName);
  if(qName.equals("map")) {
      String from=attributes.getValue("from");
      String to=attributes.getValue("to");
      if(_currentMap!=null)
        _currentMap.put(from, to);
      else
        Emit.out("Error; parsing "+from+" --> "+to +" with no stdf record type defined.");
    }
    _sbValue.setLength(0); //clear out the 'value'
  }

  /**XML Parsing. Called for the tag 'value' content*/
  @Override
  public void characters(char[] ch, int start, int length) throws SAXException {
    _sbValue.append(ch, start, length);  
  }

  /**XML Parsing. Called at the end of an element*/
  @Override
  public void endElement(String uri, String localName, String qName) throws SAXException {
    if(qName.equals("name")) { //prr, mir, etc
      String stdf4Type = _sbValue.toString();
      _currentMap = _map.get(stdf4Type); //see if one already exists for this type
      if(_currentMap==null) {
        _currentMap = new HashMap<String, String>();
        _map.put(stdf4Type, _currentMap);
      }
    }
    _sbValue.setLength(0); //clear out the 'value'
    tagStack.pop();
  }

  
  /**Test main*/
  public static void main(String[] args) {
    StdfMapping xmlReader =new StdfMapping();
    HashMap<String, HashMap<String, String>> map = xmlReader._map;
    Iterator<String> iter = map.keySet().iterator();
    while(iter.hasNext()) {
      String mapName = iter.next();
      System.out.println("MAP:"+mapName);
      HashMap<String, String> amap = map.get(mapName);
      System.out.println("    >"+amap);
    }
    System.out.println("Testing getMapFor:");
    System.out.println(" mir:"+xmlReader.getMapFor("mir"));
    System.out.println(" wir:"+xmlReader.getMapFor("wir"));
    System.out.println(" mrr:"+xmlReader.getMapFor("mrr"));
  }
  
}
