package ri.stdf;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.UIManager;

import org.sqlite.SQLiteJDBCLoader;

import ri.core.rlinda.RlindaMsgTestApp;
import ri.util.AboutBox;
import ri.util.DbaseUtil;
import ri.util.DialogSqliteConnect;
import ri.util.Emit;
//for sqlite

/**Utility to upload STDF files into an SQL database.
 * */
public class StdfUpload extends JFrame {
  public final static boolean USE_SQLITE = true; //set to true for sqlite, false for postgres (or other non-sqlite dbases)
  
  private static final String _appVersionNo = "1.0"; //version number for this application
  private JFileChooser _chooserOutDir = new JFileChooser(); //chooser for output directory
  private JFileChooser _chooserInFiles = new JFileChooser(); //chooser for input files
  private boolean _inDirInitialized=false; //indicates the input directory has gotten its initial setting
  private boolean _outDirInitialized=false; //indicates the output directory has gotten its initial setting

  private JComboBox<String> _cboxTranslateType = new JComboBox<String>(new String[] {
      "1: Normal","2: Pure","2: TestByName"
  });
  private JCheckBox _chkCreateNarrow = new JCheckBox("Create Narrow Table", true);
  private JCheckBox _chkCreateWide = new JCheckBox("Create Wide Tables", false);
  private JCheckBox _chkOneDbasePerFile = new JCheckBox("One Database Per File", true);
  private JCheckBox _chkAutoDropPrevTables = new JCheckBox("Auto-drop previous tables", false);
  private JTextField _tfldOutDir = new JTextField(35);
  private JTextField _tfldDbaseName = new JTextField("default.sqlite", 20);
  private static final DecimalFormat fmt= new DecimalFormat("###,###,###,##0");
  private Connection _logDbConnection =null;
  private PreparedStatement _logDbStatement = null;
  private StdfUpload _me=this; //convenience thing
  private Connection _guruKeysDbConnection =null;
  private RlindaMsgTestApp _rlindaMsgTestApp=null;

  /**Constr*/
  public StdfUpload() {
    super("STDF Database Upload");
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    guiInit();
    setLocation(120,120);
    //setSize(600,400);
    //validate();
    pack();
    setVisible(true);
    initializeLogDatabase(false, true);
  }

  /**Initialize gui components*/
  private void guiInit() {

    setupMenuBar();
    JButton butnSelectDir = new JButton("Select"); //select output directory
    JButton butnConvertFromStdf = new JButton("Convert STDF File(s) To Database");

    Box pnlN = new Box(BoxLayout.Y_AXIS);
    
    JPanel pnlN1 = new JPanel(new BorderLayout());
    JPanel pnlN2 = new JPanel();
    JPanel pnlN3 = new JPanel();
    JPanel pnlN4 = new JPanel();
    JPanel pnlN5 = new JPanel();

    pnlN.add(pnlN1);
    pnlN.add(pnlN2);
    pnlN.add(pnlN3);
    pnlN.add(pnlN4);
    pnlN.add(pnlN5);

    pnlN1.add(new JLabel("Output Directory:"), BorderLayout.WEST);
    pnlN1.add(_tfldOutDir, BorderLayout.CENTER);
    pnlN1.add(butnSelectDir, BorderLayout.EAST);

    pnlN2.add(new JLabel("Database name:"));
    pnlN2.add(_tfldDbaseName);
    pnlN2.add(_chkOneDbasePerFile);
    pnlN2.add(_chkAutoDropPrevTables);

    pnlN3.add(_chkCreateWide);
    pnlN3.add(_chkCreateNarrow);

    pnlN4.add(new JLabel("Translation Type:"));
    pnlN4.add(_cboxTranslateType);

    pnlN5.add(butnConvertFromStdf);

    _chooserOutDir.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
    _chooserInFiles.setMultiSelectionEnabled(true);
    
    //Initialize gui to match initial settings:
    _tfldDbaseName.setEditable(!_chkOneDbasePerFile.isSelected()); //initialize to match checkbox setting
    _tfldOutDir.setText(_chooserOutDir.getCurrentDirectory().getAbsolutePath()); //let initial display reflect actual setting

    butnConvertFromStdf.setToolTipText("Converts STDF disk file(s) to database format");
    _chkCreateWide.setToolTipText("If selected then wide tables are generated");
    _chkCreateNarrow.setToolTipText("If selected then a narrow table is generated");
    _chkOneDbasePerFile.setToolTipText("If selected then one database is created for each file encountered (otherwise all input files go into a single database)");
    _chkAutoDropPrevTables.setToolTipText("Automatically drops previous existing data tables.");
    _tfldOutDir.setToolTipText("File directory to use for database output file(s)");
    butnSelectDir.setToolTipText("Select the output directory");
    _tfldDbaseName.setToolTipText("Name of output database file (when combining all input files into one database)");
    _cboxTranslateType.setToolTipText("Determines how non-standard fields are handled.");
    
    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(pnlN, BorderLayout.NORTH);
    Component emit = Emit.checkoutTextArea(true);
    if(emit!=null) {
      getContentPane().add(emit, BorderLayout.CENTER);
      Emit.setTextAreaRows(8);
    }
    
    butnConvertFromStdf.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doConvertFromStdf(); }
    });
    
    _chkOneDbasePerFile.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent arg0) {
        _tfldDbaseName.setEditable(!_chkOneDbasePerFile.isSelected()); //set to match checkbox setting
      }
    });
    
    butnSelectDir.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent arg0) {
        _chooserOutDir.rescanCurrentDirectory();
        int result = _chooserOutDir.showOpenDialog(_me);
        if(result==JFileChooser.CANCEL_OPTION) 
          return;
        _tfldOutDir.setText(_chooserOutDir.getSelectedFile().getAbsolutePath()); //let selected value reflect the actual setting
        _outDirInitialized=true; //indicate initialized so doesn't get changed by input side
        if(!_inDirInitialized) { //if the input chooser has not been used yet, initialize to same as output side
          File f = _chooserOutDir.getSelectedFile();
          if(f==null) f = _chooserOutDir.getCurrentDirectory();
          _chooserInFiles.setCurrentDirectory(f);
          _inDirInitialized=true;
        }
      }
    });
    
  }

  /**Sets up pulldown menus*/
  private void setupMenuBar() {

    JMenu menuFile=new JMenu("File");
    JMenu menuOption=new JMenu("Option");
    JMenu menuTools=new JMenu("Tools");
    JMenu menuHelp=new JMenu("Help");

    JMenuItem menuFileExit =new JMenuItem("Exit");
    JMenuItem menuOptionDropLogInfo = new JMenuItem("Clear the 'Info Log' Database");
    JMenuItem menuOptionSaveToStdf4 = new JMenuItem("Save To STDF4 file");
    JMenuItem menuOptionIndexMetadata = new JMenuItem("Index Metadata");
    JMenuItem menuToolsTuples = new JMenuItem("Tuple Queries");
    JMenuItem menuHelpAbout=new JMenuItem("About");
    menuOptionDropLogInfo.setToolTipText("Clears out the information (log) database");
    menuOptionSaveToStdf4.setToolTipText("Saves file represented in the Narrow table, to an STDF4 file on disk");

    // --------------------------------------------------
    JMenuBar menuBar=new JMenuBar();
    setJMenuBar(menuBar);
    menuBar.add(menuFile);
    menuBar.add(menuOption);
    menuBar.add(menuTools);
    menuBar.add(menuHelp);
    menuFile.add(menuFileExit);
    menuOption.add(menuOptionDropLogInfo);
    menuOption.add(menuOptionSaveToStdf4);
    menuOption.add(menuOptionIndexMetadata);
    menuTools.add(menuToolsTuples);
    menuHelp.add(menuHelpAbout);

    // --------------------------------------------------
    menuFileExit.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { System.exit(0); }
    });

    menuOptionDropLogInfo.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { initializeLogDatabase(true, true); }
    });
    menuOptionSaveToStdf4.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doSaveStdf4(); }
    });
    menuOptionIndexMetadata.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doIndexMetadata(); }
    });
    menuToolsTuples.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { 
        if(_rlindaMsgTestApp==null)
          _rlindaMsgTestApp = new RlindaMsgTestApp(false);
        _rlindaMsgTestApp.setRuleExamples(getRlindaExamples());
        _rlindaMsgTestApp.setVisible(true);
      }
    });

    menuHelpAbout.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doHelpAbout(); }
    });
  }
  
  /**Returns true if tables in previous versions of the file should be dropped (removed)*/ 
  public boolean isAutodropPreviousTables() { return _chkAutoDropPrevTables.isSelected(); }
  
  //------------ 'Info Log' related methods: ------------- 
  
  /**Create the "info log" database. Optionally clears out previously data.*/
  void initializeLogDatabase(boolean clearPrevious, boolean verbose) {
    String jdbcClassName = "org.sqlite.JDBC";
    try { Class.forName(jdbcClassName); } //ensure the sqlite JDBC class is available (throws ClassNotFoundException)
    catch (ClassNotFoundException e) { Emit.out("Error: unable to load the JDBC class ("+jdbcClassName+")"); }
    Class c = null;

    if(SQLiteJDBCLoader.isPureJavaMode())
      Emit.out("Sqlite running in pure java mode.");
    Emit.out("sqlite ver "+ SQLiteJDBCLoader.getMajorVersion()+ "."+SQLiteJDBCLoader.getMinorVersion());
    
    String loadInfoDbaseName = "loadInfo.sqlite";
    String connectTo = "jdbc:sqlite:"+ loadInfoDbaseName;
    if(verbose)
      Emit.out("'Info log' connection via: '"+connectTo+"' (driver="+jdbcClassName+")");

    try { if(_logDbConnection!=null) _logDbConnection.close(); } catch(Exception e1) {} //just in case is open

    try { _logDbConnection = DriverManager.getConnection(connectTo); } //open connection to 'info log' database
    catch(Exception e) { Emit.out("Error: unable to connect to '"+loadInfoDbaseName+"' database because: "+e); } 
    
    try {
      if(clearPrevious) //remove existing table
        _logDbConnection.createStatement().executeUpdate("drop table if exists narrow");
      _logDbConnection.createStatement().executeUpdate("create table if not exists narrow (entityID INTEGER,name TEXT,value TEXT);");
    }
    catch (SQLException e) {
      System.out.print(" Unable to clear 'info log' database narrow table because: "+e);
      Emit.out("Error creating 'info log' narrow table: "+e);
    }   

    try { _logDbStatement=_logDbConnection.prepareStatement("insert into narrow values (?,?,?);"); }
    catch (SQLException e) {
      Emit.out("Error creating loadInfo prepared statement: "+e);
    }
    if(verbose) {
      if(clearPrevious) Emit.out("Cleared 'info log' table");
      else Emit.out("Connected to 'info log' table");
    }
  }

  /**Closes up the "info log" database*/
  public void closeLogDbConnection() {
    try {
      if(_logDbStatement!=null)
        _logDbStatement.close();
    }
    catch(Exception e) {}
    try { 
      if(_logDbConnection!=null) 
        _logDbConnection.close(); 
    } 
    catch(Exception e1) {} //just in case is open
    _logDbStatement=null;
    _logDbConnection=null;
    
  }
  
  /**Returns the highest entity id found in the 'info log' database*/
  public int getLastEntityId() {
    if(_logDbConnection==null) //open info database if is currently closed
      initializeLogDatabase(false, false);
    
    ResultSet results = null;
    try {
      results=_logDbConnection.createStatement().executeQuery("SELECT MAX(entityID) FROM narrow" );
      results.next();
      return results.getInt(1);
    }
    catch (SQLException e) {
      Emit.out("Error fetching last entity ID: "+e.getMessage());
      return 0;
    }
  }
  
  /**Appends info to the 'info log' database
   * Parameters: 'val' can be String, Long, Double
   * */
  public boolean insertLog(int eid, String name, Object val) {
    if(_logDbConnection==null) //open info database if is currently closed
      initializeLogDatabase(false, false);

    PreparedStatement prep = _logDbStatement;
    try {
      prep.setInt(1, eid);
      prep.setString(2, name);     
      if(val != null) {
        Class clazz = val.getClass();
        if     (clazz == String.class)  prep.setString(3, (String)val);   
        else if(clazz == Long.class)    prep.setLong(3, (Long)val);   
        else if(clazz == Integer.class) prep.setInt(3, (Integer)val);   
        else if(clazz == Double.class)  prep.setDouble(3, (Double)val);   
        else if(clazz == Float.class)   prep.setFloat(3, (Float)val);
        //else if(clazz == byte[])      prep.setBytes(5, (byte[])val);}
        else                            prep.setString(3, val.toString()); //catch-all
      }
      prep.addBatch();
      prep.executeBatch();
    }
    catch(Exception e) {
      String emsg = "Error in "+getClass().getName()+".insertObject(): "+e; 
      Emit.out("Error inserting eid="+eid+", name="+name+", val="+val+": "+e);
    }
    return true;
  }

  // -------------------------
  
  public String getDbaseNameField() { return _tfldDbaseName.getText(); }
  
  public boolean isOneDbasePerFile() { return _chkOneDbasePerFile.isSelected(); }
  
  public String getOutDir() { 
    String outDir = _tfldOutDir.getText();
    if(outDir.endsWith("/") || outDir.endsWith("\\"))
      return outDir;
    return outDir+System.getProperty("file.separator"); //i.e. slash or backslash
  }

  /**Returns the user selected 'translation type'*/
  public String getTranslationType() {
    int ndx = _cboxTranslateType.getSelectedIndex();
    if(ndx<0) return null;
    return _cboxTranslateType.getItemAt(ndx); 
  }

  /**Drop all tables in the database*/
//void doDropTables() {
//  DialogSqliteConnect dc = DialogSqliteConnect.instance();
//  if(dc!=null) {
//    try {
//      Connection sqlConnection = dc.getConnection();
//      Statement sqlStatement = sqlConnection.createStatement();
//      
//      DatabaseMetaData dbmd = sqlConnection.getMetaData();
//      ResultSet rs = dbmd.getTables(dc.getDbaseName(), null, null, new String[]{"TABLE"});
//      ArrayList list = new ArrayList();
//      while(rs.next())
//        list.add(rs.getString(3)); //table name is index #3
//      rs.close();
//      Iterator iter = list.iterator();
//      while(iter.hasNext()) {
//        String tablename = iter.next().toString();
//        sqlStatement.executeUpdate("drop table "+tablename);
//        Emit.out("Dropped table "+tablename);
//      }
//      sqlStatement.executeUpdate("VACUUM");
//    }
//    catch(Exception e) {
//      Emit.out("Error dropping tables: "+e);
//    }
//  }
//
//  //TODO drop the info table as well
//  
//  try {
//    _logDbConnection.close();
//  }
//  catch (SQLException e1) {
//    Emit.out("Error closing the 'info log' database: "+e1);
//  }
//  
//  createLogDatabase();
//   try {
//    _logDbConnection.createStatement().executeUpdate("drop table if exists "+ "narrow");
//    _logDbConnection.createStatement().executeUpdate(
//        "create table if not exists " + "narrow" + " ("+
//        "entityID INTEGER,"+
//        "name TEXT,"+
//        "value TEXT);"
//        );
//   }
//   catch (SQLException e) {
//     Emit.out("Error dropping the previous info table");
//   }
//  Emit.out("Dropped info table ");
//  
//}


  /**Drops all tables from the specified database (note: opens and closes its own dbase connection)*/
  void dropDataTablesFrom(File dbFile) {
    Connection sqlConnection = null;
      try {
        if(USE_SQLITE)
          sqlConnection = DbaseUtil.doConnectToSqlite(dbFile.getAbsolutePath(), false);
        else //postgres
          sqlConnection = DbaseUtil.doConnectTo("postgresql", "localhost", false, dbFile.getName(), "postgres", "rolop", true); //OER EXP

        Statement sqlStatement = sqlConnection.createStatement();
        
        DatabaseMetaData dbmd = sqlConnection.getMetaData();
        ResultSet rs = dbmd.getTables(dbFile.getName(), null, null, new String[]{"TABLE"});
        ArrayList list = new ArrayList();
        while(rs.next())
          list.add(rs.getString(3)); //table name is index #3
        rs.close();
        Iterator iter = list.iterator();
        while(iter.hasNext()) {
          String tablename = iter.next().toString();
          if(tablename!=null && !tablename.startsWith("pg_")) { //do not delete internal postgres tables
            sqlStatement.executeUpdate("drop table "+tablename);
            Emit.out("Dropped table "+tablename+" from "+dbFile.getName());
          }
        }
        sqlStatement.executeUpdate("VACUUM");
      }
      catch(Exception e) {
        Emit.out("Error dropping tables from "+dbFile.getAbsolutePath()+": "+e);
      }
      finally {
        try { sqlConnection.close(); } catch(Exception e) {}
      }
  }
  
  /**Convert an STDF4 file to database*/
  private void doConvertFromStdf() {

    _chooserInFiles.setDialogTitle("SELECT STDF FILE TO READ IN");
    _chooserInFiles.rescanCurrentDirectory();
    int response = _chooserInFiles.showOpenDialog(this); //prompt user for file name(s)
    File[] files = _chooserInFiles.getSelectedFiles();
    if(response==JFileChooser.CANCEL_OPTION || files==null) {
      Emit.out("Canceled");
      return;
    }

    _inDirInitialized=true; //indicate initialized so doesn't get changed by output side
    if(!_outDirInitialized) { //if the output chooser has not been used yet, initialize to same as input side
      File f = _chooserInFiles.getSelectedFile();
      if(f==null) f = _chooserInFiles.getCurrentDirectory();
      else f=f.getParentFile(); //so is directory not file
      _chooserOutDir.setCurrentDirectory(f);
      _tfldOutDir.setText(f.getAbsolutePath()); //let display reflect the setting
      _outDirInitialized=true;
    }

    new StdfUploadThread(this, files).start(); //run the upload(s) from a separate thread in case it takes awhile
  }
  
  /**Prmpts for a (narrow table) database and saves its contents to a file in traditional STDF4 format.*/
  private void doSaveStdf4() {
    _chooserInFiles.setDialogTitle("SELECT SQLITE FILE TO READ IN");
    _chooserInFiles.rescanCurrentDirectory();
    int response = _chooserInFiles.showOpenDialog(this); //prompt user for sqlite file name to read in
    File dbFile = _chooserInFiles.getSelectedFile();
    if(response==JFileChooser.CANCEL_OPTION || dbFile==null) {
      Emit.out("Canceled");
      return;
    }

    _chooserInFiles.setDialogTitle("SELECT STDF FILE TO WRITE OUT");
    _chooserInFiles.setSelectedFile(new File("")); //clear entry selected in the previous prompt
    _chooserInFiles.rescanCurrentDirectory();
    response = _chooserInFiles.showSaveDialog(this); //prompt user for file name to save to
    File stdfFile = _chooserInFiles.getSelectedFile();
    if(response==JFileChooser.CANCEL_OPTION || stdfFile==null) {
      Emit.out("Canceled");
      return;
    }

    Connection sqlConnection = null;
    try {
      new SaveStdf4Thread(this, dbFile, stdfFile).start(); //run the save from a separate thread in case it takes awhile
    }
    catch(Exception e) {
      Emit.out("Unable to convert "+dbFile.getAbsolutePath()+" to STDF4 file "+stdfFile.getAbsolutePath()+" because: "+e);
    }
  }
  
  /** */
  public boolean isCreateWideSelected() {
    return _chkCreateWide.isSelected();
  }

  /** */
  public boolean isCreateNarrowSelected() {
    return _chkCreateNarrow.isSelected();
  }


  /**Utility function to return the number of rows in a database.
   * note: expects an already open SQL Connection to database.
   * */
  public int getDbRowCount(Connection sqlConnection, File dbFile) {
    
    int totalRecordCnt=0;
    try {
      if(sqlConnection==null || sqlConnection.isClosed() || dbFile==null) return 0;
      
      DatabaseMetaData md = sqlConnection.getMetaData(); //get info on database tables
      ResultSet rs = md.getTables(dbFile.getName(), null, null, new String[]{"TABLE"});
      ArrayList<String> tableNames = new ArrayList<String>(); //list of tables in the database
      while(rs.next())
        tableNames.add(rs.getString(3)); //table name is index #3
      rs.close();

      //Get info on database records:
      totalRecordCnt=0; //total number of records in all the tables
      Iterator iter = tableNames.iterator();
      for(String tableName : tableNames) {
        int recordCount=0;
        ResultSet results = sqlConnection.createStatement().executeQuery("SELECT COUNT(*) FROM "+tableName);
        if(results.next())
          recordCount = results.getInt(1);
        //Emit.out("Table "+tableName+" has "+recordCount+" records"); //debug
        totalRecordCnt += recordCount;
        results.close();
      }
      
    }
    catch(Exception e) {
      Emit.out("Unable to get number of database rows from '"+dbFile.getAbsolutePath()+"' because: "+e.getMessage());
      return 0;
    }

    return totalRecordCnt;
  }
  
  /**Displays the Help | About Dialog*/
  private void doHelpAbout() {
    new AboutBox(this, new String[] {
        "STDF Upload Tool",
        "Version: " + _appVersionNo,
        //"Copyright Roos Instruments (c) "+AboutBox.getYearStringFor(2013),
        ""
        }, null);
  }
  
  /** reads files in directory and generates the .gat file and copies the
   * body file to one with sha1 hash as name.
   */
  private void doCreateMetadataOld(){
    Emit.out("Creating metadata files");
    new RitdbMetadataCreator(new File(getOutDir()));
  }
  private void doIndexMetadata(){
    new RitdbIndexingThread(this).start(); //run the upload(s) from a separate thread in case it takes awhile
  }
  
  /** generates and then reads .gat files in directory and updates index database
   */
  private void doIndexMetadataOld(){

    long t0 = System.currentTimeMillis();
    new RitdbMetadataCreator(new File(getOutDir())); // generate the files
    long t1 = System.currentTimeMillis();
    Emit.out("  Created Metadata ("+ ((t1-t0)/1000.0) +" sec)\n");
    Emit.out("Indexing metadata files");
    // need to open or create the indexDB
    Connection db =null;
    String dbName = "guruKeys.sqlite";
    try { db = DriverManager.getConnection("jdbc:sqlite:"+dbName); } //open connection to 'guruKeys' database
    catch(Exception e) { 
      Emit.out("Error. Unable to connect to database "+dbName+" because: "+e);
      return;
    } 
    try {
      db.createStatement().executeUpdate(
          "create table if not exists " + "narrow" + " ("+
          "sequence INTEGER PRIMARY KEY,"+
          "entityID INTEGER,"+
          "indexID INTEGER,"+
          "name TEXT,"+
          //"value TEXT,"+ //for postgres
          "value NONE,"+
          "value2 NONE);"
          );
    }
    catch (SQLException e1) {
      // TODO Auto-generated catch block
      e1.printStackTrace();
    }
    try { 
      db.setAutoCommit(false);
      new MetadataIndexer(new File(getOutDir()), db); 
      db.commit(); //commit all those changes
      db.setAutoCommit(true);
    }
    catch (SQLException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
    finally { //ensures the database gets closed regardless of exit path
      try { db.close(); } catch(Exception e) {}
    }
    t0 = System.currentTimeMillis();
    Emit.out("  Finished Indexing Metadata ("+ ((t0-t1)/1000.0) +" sec)\n");
  }
  
  private HashMap getRlindaExamples(){
      HashMap m = new HashMap(); //any sort of a ‘map’ is fine
      m.put("MIR", 
              "find=~set1~?name~?value\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value\n"+ 
              "rule=~set1~narrow~1~_~?name~?value\n"+
              "function=~set1~sort~ASC~name"
          );
      m.put("Entities", 
              "find=~set1~?value~?eid\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value\n"+ 
              "rule=~set1~narrow~?eid~_~ENTITY_TYPE~?value\n"+
              "function=~set1~sort~ASC~value"
          );
      m.put("Test info 2", 
              "find=~set1~?name~?value\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value\n"+ 
              "rule=~set1~narrow~?eid~_~ENTITY_TYPE~TEST_INFO\n"+
              "rule=~set1~narrow~?eid~_~TEST_NUMBER~2\n"+
              "rule=~set1~narrow~?eid~_~?name~?value\n"+
              "function=~set1~sort~ASC~value"
          );
      m.put("Test names", 
              "find=~set1~?name\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value\n"+ 
              "rule=~set1~narrow~?eid~_~ENTITY_TYPE~TEST_INFO\n"+
              "rule=~set1~narrow~?eid~_~TEST_NAME~?name\n"+
              "function=~set1~sort~ASC~name"
          );
      
      m.put("Test result 2", 
              "find=~set1~?value~?pf\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value~narrow.value2\n"+ 
              "rule=~set1~narrow~?tid~_~ENTITY_TYPE~TEST_INFO~_\n"+
              "rule=~set1~narrow~?tid~_~TEST_NUMBER~2~_\n"+
              "rule=~set1~narrow~?eid~_~ENTITY_TYPE~TEST_EVENT~_\n"+
              "rule=~set1~narrow~?eid~?tid~_~?value~?pf\n"+
              "function=~set1~sort~ASC~value"
          );
      
      m.put("Test info", 
              "find=~set1~?name~?index~?number~?units~?hlim~?llim\n"+ 
              "tuple=~narrow~narrow.entityID~narrow.indexID~narrow.name~narrow.value~narrow.value2\n"+ 
              "rule=~col1~narrow~?event~_~ENTITY_TYPE~TEST_INFO~_\n"+
              "rule=~col1~narrow~?event~0~TEST_NUMBER~?number~_\n"+
              "rule=~col1~narrow~?event~0~TEST_NAME~?name~_\n"+
              "rule=~col1~narrow~?event~0~TEST_UNIQUE_ID~?index~_\n"+
              "rule=~col1~narrow~?event~0~TEST_UNITS~?units~_\n"+
              "rule=~col1~narrow~?limits~0~ENTITY_TYPE~TEST_LIMITS~_\n"+
              "rule=~col1~narrow~?limits~?event~UL~?hlim~_\n"+
              "rule=~col1~narrow~?limits~?event~LL~?llim~_\n"+
              "function=~col1~nullable~hlim~llim\n"+
              "function=~col1~sort~ASC~name"
          );
      return m;
  }
  
  /**Main*/
  public static void main(String[] args) {
    try { UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); } catch(Exception e) {}
    new StdfUpload();
  }

}
