package ri.stdf;

import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;

import ri.util.Emit;
import ri.util.DialogSqliteConnect;

/**Performs uploading of STDF files, in a separate thread*/
public class StdfUploadThread extends Thread {
  private File[] _files;      //list of files to import
  private StdfUpload _parent; //parent window, for settings and display purposes
  
  /**Constr*/
  public StdfUploadThread(StdfUpload parent, File[] files) {
    _parent = parent;
    _files=files;
  }
  
  @Override
  public void run() {
      long t0 = System.currentTimeMillis();
      int entity = _parent.getLastEntityId(); //used for database insert of file info
      for(int i=0; i < _files.length; i++) {  //upload each of the selected files
        entity++;
        File file = _files[i];
        String filename = file.getName();

        String dbName; //get or create name of database to be used for this file
        if(_parent.isOneDbasePerFile()) {
          dbName = filename;
          int p = dbName.lastIndexOf(".");
          if(p>=0) dbName = dbName.substring(0, p);
          dbName+=".ritdb";
        }
        else dbName = _parent.getDbaseNameField();
        File dbFile = new File(_parent.getOutDir()+dbName); //create 'File' object from complete pathname
        
        Emit.out("#"+(i+1)+": "+filename+" ==> "+dbFile.getAbsolutePath());
        
        if(_parent.isAutodropPreviousTables()) {
          _parent.dropDataTablesFrom(dbFile);
          Emit.out("  Removed previous data table(s).");
        }
        
        _parent.insertLog(entity, "fileName", filename);
        long fileSize=new File(filename).length(); //size of stdf  file on disk
        _parent.insertLog(entity, "stdfFileSize", file.length());
        //Emit.out("\n");
        Connection sqlConnection = null; //connection to database
        try {
          if(StdfUpload.USE_SQLITE)
            sqlConnection = DialogSqliteConnect.doConnectTo(dbFile.getAbsolutePath(), false);
          else //postgres
            //sqlConnection = DialogConnect.doConnectTo("postgresql", "localhost", false, dbFile.getName(), "postgres", "rolop", true); //OER EXP
          if(sqlConnection==null)  
            throw new Exception("No connection available to database '"+dbFile.getAbsolutePath()+"'");
          
          if(_parent.isCreateWideSelected()) {
            long t1 = System.currentTimeMillis();
            RiStdfParser sc = new RiStdfParser(file, sqlConnection);
            long t2 = System.currentTimeMillis();
            Emit.out("  Inserted STDF records into wide tables ("+ ((t2-t1)/1000.0) +" sec)");
          }
          
          if(_parent.isCreateNarrowSelected()) {
            long t3 = System.currentTimeMillis();
            Stdf4ToStdf5Translator sc2 = new Stdf4ToStdf5Translator(file, sqlConnection, _parent.getTranslationType());
            long t4 = System.currentTimeMillis();
            Emit.out("  Inserted STDF records into narrow table ("+ ((t4-t3)/1000.0) +" sec)\n");
            _parent.insertLog(entity, "stdfRecords", sc2.recordCnt);
            _parent.insertLog(entity, "databaseSize", dbFile.length());
            _parent.insertLog(entity, "databaseRecordCount", _parent.getDbRowCount(sqlConnection, dbFile));
            _parent.insertLog(entity, "conversionTime", ((t4-t3)/1000.0));
          }
        }
        catch(SQLException e) {
          Emit.out("Unable to connect to database '"+dbFile.getAbsolutePath()+"' because: "+e);
        }
        catch(IOException e) {
          String msg = "File Error with ["+file.getAbsolutePath()+"]: "+e;
          Emit.out(msg);
          System.out.println(msg);
        }
        catch(Exception e) {
          String msg = "Error uploading STDF file "+file.getName()+": "+e; 
          Emit.out(msg);
          System.out.println(msg);
          e.printStackTrace();
        }
        finally { //ensure dbase connection gets closed each time
          try { if(sqlConnection!=null) sqlConnection.close(); }
          catch(Exception e) { Emit.out("Error closing database '"+dbName+"': "+e); }
          _parent.closeLogDbConnection(); //close up the 'info log' database (note: auto-reopens when needed)
        }
      }
    long t5 = System.currentTimeMillis();
    Emit.out("\nTotal time for all files: "+ ((t5-t0)/1000.0) +" sec)");
  }

}
