package ri.stdf;
import java.io.File;
import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;

import org.sqlite.SQLiteJDBCLoader;

public class Translate {
  private File  _outDir;
  private ArrayList<File> _fileList;      //list of files to import
  private File[] _files;
  private Connection _logDbConnection =null;
  private PreparedStatement _logDbStatement = null;
  private Connection _guruKeysDbConnection =null;
  private int _logId;  // the current log id
  
  /**Constr*/
  public Translate(String args[]) {
    // check the output directory
    _outDir = new File(args[0]);
    if(!(_outDir.exists())){
      System.out.println(_outDir+" does not exist");
      return;
    }
    if(!(_outDir.isDirectory())){
      System.out.println(_outDir+" is not a directory");
      return;
    }
    // got this far so set up the database
    _logId = getLastEntityId(); 
    // create s list of files, expands directories if present
    // expects a file or a directory followed by a filter
    File tmpFile = null;
    for( int i = 1; i< args.length; i++){
      tmpFile = new File(args[i]);   
      if(tmpFile.isDirectory()){
        System.out.println("expanding dir "+tmpFile.getPath());
        DirectoryStream<Path> stm = null;
        try {
          stm = Files.newDirectoryStream(tmpFile.toPath(), args[i + 1]);
          i++;
          Iterator<Path> stmnext = stm.iterator();
          while(stmnext.hasNext()){
            translate(stmnext.next().toFile());
          }
        }
        catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }

        //File[] files = tmpFile.listFiles((FilenameFilter) new WildcardFilter("?e*.t?t"));

        //for( int j= 0; j < files.length; j++){
         // translate(files[j]);
        //}
      }else{
        translate(tmpFile);
      }
    }
    closeLogDbConnection(); //close up the 'info log' database (note: auto-reopens when needed)
  }
  
  private void translate(File file){
  long t0 = System.currentTimeMillis();
    _logId++;
    String filename = file.getName();
    String dbName; //get or create name of database to be used for this file
    dbName = filename;
    int p = dbName.lastIndexOf(".");
    if(p>=0) dbName = dbName.substring(0, p);
    dbName+=".ritdb";
    File dbFile = new File(_outDir, dbName); //create 'File' object from complete pathname
    if(dbFile.exists()){
      System.out.println(file.getAbsolutePath() + " output file exists");
      return;}
    System.out.println(filename+" ==> "+dbFile.getAbsolutePath());
    insertLog(_logId, "fileName", filename);
    long fileSize=new File(filename).length(); //size of stdf  file on disk
    insertLog(_logId, "stdfFileSize", file.length());
    //System.out.println("create db");
    Connection sqlConnection = null; //connection to database
    try {
      sqlConnection = doConnectToSqlite(dbFile.getAbsolutePath());
      if(sqlConnection==null)  
        throw new Exception("No connection available to database '"+dbFile.getAbsolutePath()+"'");
      //System.out.println("start translate");
      long t3 = System.currentTimeMillis();
      Stdf4ToStdf5Translator sc2 = new Stdf4ToStdf5Translator(file, sqlConnection, "normal");
      long t4 = System.currentTimeMillis();
      System.out.print("  ("+ ((t4-t3)/1000.0) +" sec)\n");
      insertLog(_logId, "stdfRecords", sc2.recordCnt);
      insertLog(_logId, "databaseSize", dbFile.length());
      insertLog(_logId, "databaseRecordCount", getDbRowCount(sqlConnection, dbFile));
      insertLog(_logId, "conversionTime", ((t4-t3)/1000.0));
    }
    catch(SQLException e) {
      System.out.println("Unable to connect to database '"+dbFile.getAbsolutePath()+"' because: "+e);
    }
    catch(IOException e) {
      String msg = "File Error with ["+file.getAbsolutePath()+"]: "+e;
      System.out.println(msg);
    }
    catch(Exception e) {
      String msg = "Error uploading STDF file "+file.getName()+": "+e; 
      System.out.println(msg);
      e.printStackTrace();
    }
    finally { //ensure dbase connection gets closed each time
      try { if(sqlConnection!=null) sqlConnection.close(); }
      catch(Exception e) { System.out.println("Error closing database '"+dbName+"': "+e); }
  }
long t5 = System.currentTimeMillis();

}
  
  
  /**Appends info to the 'info log' database
   * Parameters: 'val' can be String, Long, Double
   * */
  public boolean insertLog(int eid, String name, Object val) {
    if(_logDbConnection==null) //open info database if is currently closed
      initializeLogDatabase();
    PreparedStatement prep = _logDbStatement;
    try {
      prep.setInt(1, eid);
      prep.setString(2, name);     
      if(val != null) {
        Class clazz = val.getClass();
        if     (clazz == String.class)  prep.setString(3, (String)val);   
        else if(clazz == Long.class)    prep.setLong(3, (Long)val);   
        else if(clazz == Integer.class) prep.setInt(3, (Integer)val);   
        else if(clazz == Double.class)  prep.setDouble(3, (Double)val);   
        else if(clazz == Float.class)   prep.setFloat(3, (Float)val);
        //else if(clazz == byte[])      prep.setBytes(5, (byte[])val);}
        else                            prep.setString(3, val.toString()); //catch-all
      }
      prep.addBatch();
      prep.executeBatch();
    }
    catch(Exception e) {
      String emsg = "Error in "+getClass().getName()+".insertObject(): "+e; 
      System.out.println("Error inserting eid="+eid+", name="+name+", val="+val+": "+e);
    }
    return true;
  }
  
  /**Closes up the "info log" database*/
  public void closeLogDbConnection() {
    try {
      if(_logDbStatement!=null)
        _logDbStatement.close();
    }
    catch(Exception e) {}
    try { 
      if(_logDbConnection!=null) 
        _logDbConnection.close(); 
    } 
    catch(Exception e1) {} //just in case is open
    _logDbStatement=null;
    _logDbConnection=null;
    
  }
  /**Create the "info log" database*/
  void initializeLogDatabase() {
    String jdbcClassName = "org.sqlite.JDBC";
    try { Class.forName(jdbcClassName); } //ensure the sqlite JDBC class is available (throws ClassNotFoundException)
    catch (ClassNotFoundException e) { System.out.println("Error: unable to load the JDBC class ("+jdbcClassName+")"); }
    Class c = null;

    if(SQLiteJDBCLoader.isPureJavaMode())
      System.out.println("Sqlite running in pure java mode.");
    System.out.println("sqlite ver "+ SQLiteJDBCLoader.getMajorVersion()+ "."+SQLiteJDBCLoader.getMinorVersion());
    //System.out.println("connecting to SQLite driver");
    String loadInfoDbaseName = "loadInfo.sqlite";
    String connectTo = "jdbc:sqlite:"+ loadInfoDbaseName;
    System.out.println("'Info log' connection via: '"+connectTo+"' (driver="+jdbcClassName+")");
    //System.out.println("close if open");
    try { if(_logDbConnection!=null) _logDbConnection.close(); } catch(Exception e1) {} //just in case is open
    //System.out.println("get conn");
    try { _logDbConnection = DriverManager.getConnection(connectTo); } //open connection to 'info log' database
    catch(Exception e) { System.out.println("Error: unable to connect to '"+loadInfoDbaseName+"' database because: "+e); } 
    //System.out.println("create table");
    try {
      _logDbConnection.createStatement().executeUpdate("create table if not exists narrow (entityID INTEGER,name TEXT,value TEXT);");
    }
    catch (SQLException e) {
      System.out.print(" Unable to clear 'info log' database narrow table because: "+e);
      System.out.println("Error creating 'info log' narrow table: "+e);
    }   
    System.out.println("prepare stat");
    try { _logDbStatement=_logDbConnection.prepareStatement("insert into narrow values (?,?,?);"); }
    catch (SQLException e) {
      System.out.println("Error creating loadInfo prepared statement: "+e);
    }
    System.out.println("Connected to 'info log' table");
  }
  
  /**Utility function to return the number of rows in a database.
   * note: expects an already open SQL Connection to database.
   * */
  public int getDbRowCount(Connection sqlConnection, File dbFile) {
    
    int totalRecordCnt=0;
    try {
      if(sqlConnection==null || sqlConnection.isClosed() || dbFile==null) return 0;
      
      DatabaseMetaData md = sqlConnection.getMetaData(); //get info on database tables
      ResultSet rs = md.getTables(dbFile.getName(), null, null, new String[]{"TABLE"});
      ArrayList<String> tableNames = new ArrayList<String>(); //list of tables in the database
      while(rs.next())
        tableNames.add(rs.getString(3)); //table name is index #3
      rs.close();

      //Get info on database records:
      totalRecordCnt=0; //total number of records in all the tables
      Iterator iter = tableNames.iterator();
      for(String tableName : tableNames) {
        int recordCount=0;
        ResultSet results = sqlConnection.createStatement().executeQuery("SELECT COUNT(*) FROM "+tableName);
        if(results.next())
          recordCount = results.getInt(1);
        //Emit.out("Table "+tableName+" has "+recordCount+" records"); //debug
        totalRecordCnt += recordCount;
        results.close();
      }
      
    }
    catch(Exception e) {
      System.out.println("Unable to get number of database rows from '"+dbFile.getAbsolutePath()+"' because: "+e.getMessage());
      return 0;
    }

    return totalRecordCnt;
  }
  /**Returns the highest entity id found in the 'info log' database*/
  public int getLastEntityId() {
    if(_logDbConnection==null) //open info database if is currently closed
      initializeLogDatabase();
    
    ResultSet results = null;
    try {
      results=_logDbConnection.createStatement().executeQuery("SELECT MAX(entityID) FROM narrow" );
      results.next();
      return results.getInt(1);
    }
    catch (SQLException e) {
      System.out.println("Error fetching last entity ID: "+e.getMessage());
      return 0;
    }
  }
  /**Utility function to connect to an SQLite database
   * Throws ClassCastException if the JDBC driver class is not available
   * Throws SQLException if fails in establishing a connection to the specified database
   * */
  public static Connection doConnectToSqlite(String dbName) throws SQLException, ClassNotFoundException {
    dbName = (dbName==null) ? "" : dbName.trim();

    String jdbcClassName = "org.sqlite.JDBC";
    Class.forName(jdbcClassName); //ensure the sqlite JDBC class is available (throws ClassNotFoundException) 
    String connectTo = "jdbc:sqlite:"+dbName;
    //System.out.println("Connect using: '"+connectTo+"' (driver="+jdbcClassName+")");
    return DriverManager.getConnection(connectTo); //throws SQLException
  }

  public static void main(String[] args){
    // args will have destinationDirectory, source files regex
    String[] newArgs = new String[2];
    newArgs[0] = "/Users/markroos/stdf/loadertest/";
    newArgs[1] = "/Users/markroos/stdf/kl/STDF/AFT20P140-4WN_BEGINNING_CHECK_90_UNITS_20131020225154.stdf";
    new Translate(args); //run the upload(s) from a separate thread in case it takes awhile

  }


}
