package ri.util;

import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.util.Calendar;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;

/**General purpose About Box.  Displays a graphic and some text description.
 *  Give a string[] for the text strings and a swing.ImageIcon for the graphic.
 *  If no icon given, displays an RI Logo
 */
public class AboutBox extends JDialog {
  private Component _parent; //used in positioning the dialog
  
  /**Constr: Given an array of strings
   * (e.g: Program name, version, copyright, description...)
   * and an optional icon. If icon==null then a default RI logo is used)
   */
  public AboutBox(Frame parent, Object[] info, ImageIcon icon) {
    super(parent, "About", true);
    _parent=parent;
    final Object[] theInfo=info;
    final ImageIcon theIcon=icon;
    SwingUtilities.invokeLater(new Runnable() {
      public void run() { initAboutBox(theInfo, theIcon); }
    });
  }

  public AboutBox(Dialog parent, Object[] info, ImageIcon icon) {
    super(parent, "About", true);
    _parent=parent;
    final Object[] theInfo=info;
    final ImageIcon theIcon=icon;
    SwingUtilities.invokeLater(new Runnable() {
      public void run() { initAboutBox(theInfo, theIcon); }
    });
  }
  
  /**Given a year, returns a string that represents the interval from that year to present.
   * Examples given input of 2001: 
   *   during 2001 would return "2001", 
   *   during 2009 would return "2001 - 2009", 
   *   during 1999 would return "2001"
   * */
  public static String getYearStringFor(int givenYr) {
    int currentYr = Calendar.getInstance().get(Calendar.YEAR);
    if(currentYr > givenYr) 
      return ""+givenYr+" - "+currentYr; //i.e. "2001 - 2009"
    return Integer.toString(givenYr); //i.e. "2001"
  }

  private void initAboutBox(Object[] theInfo, ImageIcon theIcon) {
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    guiInit(theInfo, theIcon);
    pack();
    if(_parent==null) { //center the dialog
      Dimension scrnSz = Toolkit.getDefaultToolkit().getScreenSize();
      int xpos = (scrnSz.width - getSize().width) / 2;
      int ypos = (scrnSz.height - getSize().height) / 2;
      setLocation(xpos, ypos);
    }
    else setLocationRelativeTo(_parent);
    ///setLocation(200, 200);
    setVisible(true);
  }

  private void guiInit(Object[] info, ImageIcon icon) {
    if (info == null || info.length < 1)
      info=new String[] { "<PROGRAM NAME>", "Version: <VERSION NO>",
          "Copyright Roos Instruments (c) <YEAR> ", "<PROGRAM DESCRIPTION>", };

    /*Comment out because causes warnings in JarMaker: * /
    //Use default if no icon supplied. If default not found then fail silently:
    String defaultIconName="RiLogo1_5in.gif"; //Large
    if(icon==null) { 
      try { icon=new ImageIcon(getClass().getResource("/"+defaultIconName) ); } 
      catch (Exception e) { icon=null; }
    }
    if(icon == null) {
      try { icon=new ImageIcon(defaultIconName); } 
      catch (Exception e) { icon=null; }
    }
    /* */

    JPanel pnlN=new JPanel(new BorderLayout());
    JPanel pnlS=new JPanel();
    JPanel pnlImage=new JPanel();
    JPanel pnlInfo=new JPanel(new GridLayout(info.length, 1));
    JButton butn=new JButton("OK");
    setResizable(false);
    pnlImage.setBorder(new EmptyBorder(10, 10, 10, 10));
    pnlInfo.setBorder(new EmptyBorder(10, 60, 10, 10));
    if (icon != null) pnlImage.add(new JLabel(icon));
    int length=info.length;
    for (int i=0; i < length; i++)
      pnlInfo.add(new JLabel(info[i]==null ? "" : info[i].toString()), null);

    pnlN.add(pnlImage, BorderLayout.WEST);
    pnlN.add(pnlInfo, BorderLayout.CENTER);
    pnlS.add(butn);
    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(pnlN, BorderLayout.NORTH);
    getContentPane().add(pnlS, BorderLayout.SOUTH);

    butn.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { dispose(); }
    });
  }  

  /**Override to dispose of window on close*/ 
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID() == WindowEvent.WINDOW_CLOSING)
      dispose();
  }

}
