package ri.util;

import java.awt.event.MouseEvent;

/**A listener for ActionList.  Implementors of this interface 
 * provide the names of the actions as well as their implementation.
 * Creation date: (9/21/2003 12:06:21 AM)
 * @author: OER
 * */
public interface ActionListListener {

	/**Performs the action related to the specified index 
	 * (This corresponds to the list provided by getActionNames)
	 * Note: Required for interface ActionListListener
	*/
	public void doAction(int index, MouseEvent e);
	
	/**Returns an array of actions.There are two ways to use this:
	 *
	 * 1) The returned value is an array of ActionListener objects whose toString() 
	 *    method has been overridden to display the appropriate action name. 
	 *    In this case the actionPerformed() method of each of these ActionListeners 
	 *    will be called whenever that particular action is triggered. 
	 *    In this case the doAction() method is superfluous and is not used.
   *
	 * 2) Otherwise the list of actions is taken simply as an array of 'action names'.  
	 *    Whenever a particular action is triggered, the doAction() method is called 
	 *    along with the index (in the 'actions' list) of the triggering action.
	 *    
   * Note: The two ways can be freely intermixed, even within the same ActionList
   * Note: Sub-menus are accomplished by placing another ActionList in the getActions() 
   *       method of the parent ActionList.
	 * Note: Required for interface ActionListListener
	 * */
	public Object[] getActions();
}
