package ri.util;

import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;

/**Utilities that help in using the system clipboard
 * Creation date: (10/3/2003 4:11:39 PM)
 * @author: OER
 */
public class ClipboardUtil {

  /**Private Constr*/
  private ClipboardUtil() {}

  /**Gets the System clipboard data and returns it as a String
   * (Or null if nothing to get)
   * As of this writing 'requestor' is not actually used for anything.*/
  public static String getClipboardAsString(Object requestor) {
    try {
      Transferable t=Toolkit.getDefaultToolkit().getSystemClipboard().getContents(requestor);
      Object clipData=t.getTransferData(DataFlavor.stringFlavor);
      if (clipData != null) return clipData.toString();
    }
    catch (IOException e) {}
    catch (UnsupportedFlavorException e) {}
    catch (Exception e) {}
    return null;
  }

  /**Convenience methods to copy a String to the clipboard
   * Given a clipboard owner*/
  public static void putToClipboard(ClipboardOwner cOwner, String s) {
    Clipboard clipboard=Toolkit.getDefaultToolkit().getSystemClipboard();
    clipboard.setContents(new StringSelection(s), cOwner);
  }

  /**Convenience methods to copy a String to the clipboard.
   * Creates a clipboard owner that does nothing*/
  public static void putToClipboard(String s) {
    ClipboardOwner cOwner=new ClipboardOwner() {
      public void lostOwnership(Clipboard clipboard, Transferable contents) {}
    };
    putToClipboard(cOwner, s);
  }
}
