package ri.util;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.Connection;
import java.sql.SQLException;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;


/**Dialog for connecting to an SQLite (only) database*/
public class DialogSqliteConnect extends JDialog {
  private static DialogSqliteConnect _instance=null; //singleton instance
  private Connection _currentConnection=null; //the current connection.
  private String _currentDbName=null; //name of current connection.
  private PComboBox2 _tfldDbs = new PComboBox2(getClass(), "asdcdbs", 10, "trial.sqlite");
  private boolean _cancelled=true;

  /**Constr*/
  public DialogSqliteConnect(Frame parent) {
    super(parent, "SQLite Connect", true);
    guiInit();
    pack();
    if(parent!=null) 
      setLocationRelativeTo(parent);
    else 
      setLocation(20,200);
    _instance=this;
  }
  
  /**Singleton accessor*/
  public static DialogSqliteConnect instance() { return _instance; }
  
  private void guiInit() {
    _tfldDbs.setToolTipText("Name (i.e. filename) of SQLite database");

    JPanel pnl4=new JPanel(); 
    pnl4.add(new JLabel("Database Name:")); 
    pnl4.add(_tfldDbs);

    Box pnlC = new Box(BoxLayout.Y_AXIS);
    pnlC.add(pnl4);

    JButton butnOk = new JButton("OK");
    JButton butnCancel = new JButton("Cancel");
    JPanel pnlS = new JPanel();
    pnlS.add(butnOk);
    pnlS.add(butnCancel);

    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(pnlC, BorderLayout.CENTER);
    getContentPane().add(pnlS, BorderLayout.SOUTH);

    butnOk.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doOk(); }
    });
    butnCancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) { doCancel(); }
    });
  }
  
  private void doOk() {
    _cancelled=false;
    String dbName = _tfldDbs.getText();
    try {
      _currentConnection = doConnectTo(dbName, true);
      _currentDbName = dbName;
      Emit.out("Successful Connection to ["+dbName+']');
    }
    catch(ClassNotFoundException e) { Emit.out("**Error: JDBC driver not found: "+e); }
    catch(SQLException e) { Emit.out("Error connecting to database "+dbName+": "+e); }
    dispose();
  }
  
  
  private void doCancel() {
    _cancelled=true;
    dispose();
  }

  public boolean isCancelled() { return _cancelled; }
  
  /**Disconnect from current connection*/
  public void disconnect() {
    try {
      _currentConnection.close();
      Emit.out("Connection to database "+_currentDbName+" closed.");
    }
    catch(Exception e) {
      Emit.out("**Error closing connection "+_currentDbName+": "+e);
    }
    _currentConnection=null;
    _currentDbName=null;
  }

  /**Returns the current default connection*/
  public Connection getConnection() { return _currentConnection; }

  /**Names of databases we connect to*/
  public String getDbaseName() { return _currentDbName; }
  
  
  /**Utility function to connect to an SQLite database
   * Throws ClassCastException if the JDBC driver class is not available
   * Throws SQLException if fails in establishing a connection to the specified database
   * */
  public static Connection doConnectTo(String dbName, boolean verbose) throws SQLException, ClassNotFoundException {
    return DbaseUtil.doConnectToSqlite(dbName, verbose);
  }
}