package ri.util;

import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;

import javax.swing.JComboBox;


/**A JComboBox adapted to remember previous values selected and entered.
 Modifications are:
   It uses a ri.core.panel.PersistantComboBoxModel as its model, giving it persistence.
   It is editable by default
   It preselects its first entry (if it exists)
   It's (persistent) item list is presettable in the constructor
   It contains method getText() to allow it to be dropped into a JTextField usage.

 Note: If the initial list of values, given to the constructor, has commas that are to be treated
 as a single entry, then delimiters need to be used e.g. \"

 ISSUES TO ADDRESS: ***TO DO***
 -all sources 'share' remembered settings. Need a collection type that groups multiple PComboBoxes and other Persistent values.
 -inability to emit 'non' setting (even if clear display). Always uses last non-blank choice.

 Creation date: (5/28/02 2:50:10 PM)
 @author: OER
 */
public class PComboBox extends JComboBox {

  /**Constr: given a PersistentComboBoxModel*/
  public PComboBox(PersistentComboBoxModel model) {
    super(model);
    init();
  }

   /**Constr: given a name (for persistence).
    * Creates a persistant combo box with no limit on the number of entries held*/
  public PComboBox(String name) {
    super(new PersistentComboBoxModel(name, -1, (String[])null));
    init();
  }

  /**Constr: given a name (for persistence),
   * also given a max number of entries for persistence model
   * (Note: max=-1 implies no limit)
   */
  public PComboBox(String name, int maxEntries) {
    super(new PersistentComboBoxModel(name, maxEntries, (String[])null));
    init();
  }

  /**Constr: given a name (for persistence),
   * also given a max number of entries for persistence model
   * (Note: max=-1 implies no limit)
   * and given some preset values (set to null for none)
   */
  public PComboBox(String name, int maxEntries, String[] presets) {
    super(new PersistentComboBoxModel(name, maxEntries, presets));
    init();
  }

  /**Constr: given a name (for persistence),
   * also given a max number of entries for persistence model
   * (Note: max=-1 implies no limit)
   * and given some preset values as a CSV (comma separated) String
   */
  public PComboBox(String name, int maxEntries, String csvPresets) {
    super(new PersistentComboBoxModel(name, maxEntries, Csv.toArray(csvPresets)));
    init();
  }

  private void init() {
    setEditable(true);
    if(getItemCount() > 0) setSelectedIndex(0);
    //listener that hits enter when component no longer selected:
    addFocusListener(new FocusListener() {
      public void focusGained(FocusEvent e) {
        //System.out.println(">GotFocus"); //debug
      }
      public void focusLost(FocusEvent e) {
        //TODO: Hit enter (somehow)
        //System.out.println(">LostFocus:"+getEditor().getItem()); //debug
      }
    });
  }

  /**Convenience method: Returns the currently selected item as a String
   * Does not return null
   */
  public String getText() {
    Object ob=getSelectedItem();
    return (ob == null) ? "" : ob.toString();
  }

  /**Sets the first item in the combo box to be the given one*/
  public void setText(String s) { setFirstItem(s); }

  /**Sets the first item in the combo box to be the given one*/
  public void setFirstItem(String s) {
    if(s == null) s="";
    PersistentComboBoxModel model=(PersistentComboBoxModel)this.getModel();
    model.addItem(s, true, true);
    model.setSelectedItem(s);
  }

}
