package ri.util;

import javax.swing.JComboBox;

/**A JComboBox that remembers previous values selected and entered.
 * NOTE: this version uses the Java Preferences API 
 *       PComboBox uses disk files in the default directory.
 * 
 * Modifications from a JComboBox are:
 *   It uses a ri.core.panel.PComboBoxModel2 as its model, giving it persistence via the Java persistence API
 *   It is editable by default
 *   It preselects its first entry (if it exists)
 *   It's (persistent) item list is presettable in the constructor
 *   It contains method getText() to allow it to be dropped into a JTextField usage.
 *
 * Note: If the initial list of values, given to the constructor, has commas that are to be treated
 * as a single entry, then delimiters need to be used e.g. \"
 *
 * Creation date: derived from PComboBox on 1/10/2014
 * @author: OER
 * */
public class PComboBox2 extends JComboBox {

  /**Constr: given a PComboBoxModel2*/
  public PComboBox2(PComboBox2_Model model) {
    super(model);
    init();
  }

  /**Constr: given a name (for persistence),
   * also given a max number of entries for persistence model
   * (Note: max=-1 implies no limit)
   * and given some preset values (set to null for none)
   */
  public PComboBox2(Class clazz, String name, int maxEntries, String[] presets) {
    super(new PComboBox2_Model(clazz, name, maxEntries, presets));
    init();
  }

  /**Constr: given a name (for persistence),
   * also given a max number of entries for persistence model
   * (Note: max=-1 implies no limit)
   * and given some preset values as a CSV (comma separated) String
   */
  public PComboBox2(Class clazz, String name, int maxEntries, String csvPresets) {
    super(new PComboBox2_Model(clazz, name, maxEntries, csvPresets));
    init();
  }

  /** */
  private void init() {
    setEditable(true);
    if (getItemCount() > 0) 
      setSelectedIndex(0);
  }

  /**Convenience method: Returns the currently selected item as a String
   * Does not return null
   */
  public String getText() {
    Object ob=getSelectedItem();
    return (ob == null) ? "" : ob.toString();
  }

  /**Sets the first item in the combo box to be the given one*/
  public void setText(String s) {
    setFirstItem(s);
  }

  /**Sets the first item in the combo box to be the given one*/
  public void setFirstItem(String s) {
    if (s == null) s="";
    PComboBox2_Model model=(PComboBox2_Model)this.getModel();
    model.addItem(s, true, true);
    model.setSelectedItem(s);
  }
}
